import datetime
import logging
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, Any, Optional, cast  # noqa: UP035

import dagster._check as check
from dagster import PartitionKeyRange
from dagster._core.asset_graph_view.entity_subset import EntitySubset
from dagster._core.definitions.asset_daemon_cursor import AssetDaemonCursor
from dagster._core.definitions.asset_graph_subset import AssetGraphSubset
from dagster._core.definitions.asset_key import AssetCheckKey, EntityKey
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.backfill_policy import BackfillPolicy, BackfillPolicyType
from dagster._core.definitions.base_asset_graph import BaseAssetGraph
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
)
from dagster._core.definitions.declarative_automation.automation_condition_evaluator import (
    AutomationConditionEvaluator,
)
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionEvaluation,
)
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.run_request import RunRequest
from dagster._core.instance import DynamicPartitionsStore
from dagster._core.storage.tags import (
    ASSET_PARTITION_RANGE_END_TAG,
    ASSET_PARTITION_RANGE_START_TAG,
)

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance


class AutomationTickEvaluationContext:
    def __init__(
        self,
        evaluation_id: int,
        instance: "DagsterInstance",
        asset_graph: BaseAssetGraph,
        cursor: AssetDaemonCursor,
        materialize_run_tags: Mapping[str, str],
        observe_run_tags: Mapping[str, str],
        auto_observe_asset_keys: AbstractSet[AssetKey],
        asset_selection: AssetSelection,
        logger: logging.Logger,
        emit_backfills: bool,
        default_condition: Optional[AutomationCondition] = None,
        evaluation_time: Optional[datetime.datetime] = None,
    ):
        resolved_entity_keys = {
            entity_key
            for entity_key in (
                asset_selection.resolve(asset_graph) | asset_selection.resolve_checks(asset_graph)
            )
            if default_condition or asset_graph.get(entity_key).automation_condition is not None
        }
        self._total_keys = len(resolved_entity_keys)
        self._evaluation_id = evaluation_id
        self._evaluator = AutomationConditionEvaluator(
            entity_keys=resolved_entity_keys,
            default_condition=default_condition,
            instance=instance,
            asset_graph=asset_graph,
            emit_backfills=emit_backfills,
            cursor=cursor,
            evaluation_time=evaluation_time,
            logger=logger,
        )
        self._materialize_run_tags = materialize_run_tags
        self._observe_run_tags = observe_run_tags
        self._auto_observe_asset_keys = auto_observe_asset_keys or set()

    @property
    def cursor(self) -> AssetDaemonCursor:
        return self._evaluator.cursor

    @property
    def asset_graph(self) -> BaseAssetGraph:
        return self._evaluator.asset_graph

    @property
    def total_keys(self) -> int:
        return self._total_keys

    def _legacy_build_auto_observe_run_requests(self) -> Sequence[RunRequest]:
        current_timestamp = self._evaluator.evaluation_time.timestamp()
        assets_to_auto_observe: set[AssetKey] = set()
        for asset_key in self._auto_observe_asset_keys:
            last_observe_request_timestamp = (
                self.cursor.last_observe_request_timestamp_by_asset_key.get(asset_key)
            )
            auto_observe_interval_minutes = self.asset_graph.get(
                asset_key
            ).auto_observe_interval_minutes

            if auto_observe_interval_minutes and (
                last_observe_request_timestamp is None
                or (
                    last_observe_request_timestamp + auto_observe_interval_minutes * 60
                    < current_timestamp
                )
            ):
                assets_to_auto_observe.add(asset_key)

        # create groups of asset keys that share the same repository AND the same partitions definition
        partitions_def_and_asset_key_groups: list[Sequence[AssetKey]] = []
        for repository_asset_keys in self.asset_graph.split_entity_keys_by_repository(
            assets_to_auto_observe
        ):
            asset_keys_by_partitions_def = defaultdict(list)
            for asset_key in repository_asset_keys:
                partitions_def = self.asset_graph.get(asset_key).partitions_def
                asset_keys_by_partitions_def[partitions_def].append(asset_key)
            partitions_def_and_asset_key_groups.extend(asset_keys_by_partitions_def.values())

        return [
            RunRequest(asset_selection=list(asset_keys), tags=self._observe_run_tags)
            for asset_keys in partitions_def_and_asset_key_groups
            if len(asset_keys) > 0
        ]

    def _build_run_requests(self, entity_subsets: Sequence[EntitySubset]) -> Sequence[RunRequest]:
        return build_run_requests(
            entity_subsets=entity_subsets,
            asset_graph=self.asset_graph,
            run_tags=self._materialize_run_tags,
            emit_backfills=self._evaluator.emit_backfills,
        )

    def _get_updated_cursor(
        self, results: Iterable[AutomationResult], observe_run_requests: Iterable[RunRequest]
    ) -> AssetDaemonCursor:
        return self.cursor.with_updates(
            evaluation_id=self._evaluation_id,
            condition_cursors=[result.get_new_cursor() for result in results],
            newly_observe_requested_asset_keys=[
                asset_key
                for run_request in observe_run_requests
                for asset_key in cast(
                    Sequence[AssetKey],
                    run_request.asset_selection,  # auto-observe run requests always have asset_selection
                )
            ],
            evaluation_timestamp=self._evaluator.evaluation_time.timestamp(),
        )

    def _get_updated_evaluations(
        self, results: Iterable[AutomationResult]
    ) -> Sequence[AutomationConditionEvaluation[EntityKey]]:
        # only record evaluation results where something changed
        updated_evaluations = []
        for result in results:
            previous_cursor = self.cursor.get_previous_condition_cursor(result.key)
            if (
                previous_cursor is None
                or previous_cursor.result_value_hash != result.value_hash
                or not result.true_subset.is_empty
            ):
                updated_evaluations.append(result.serializable_evaluation)
        return updated_evaluations

    def evaluate(
        self,
    ) -> tuple[
        Sequence[RunRequest], AssetDaemonCursor, Sequence[AutomationConditionEvaluation[EntityKey]]
    ]:
        observe_run_requests = self._legacy_build_auto_observe_run_requests()
        results, entity_subsets = self._evaluator.evaluate()

        return (
            [*self._build_run_requests(entity_subsets), *observe_run_requests],
            self._get_updated_cursor(results, observe_run_requests),
            self._get_updated_evaluations(results),
        )


_PartitionsDefKeyMapping = dict[
    tuple[Optional[PartitionsDefinition], Optional[str]], set[EntityKey]
]


def _get_mapping_from_entity_subsets(
    entity_subsets: Iterable[EntitySubset], asset_graph: BaseAssetGraph
) -> _PartitionsDefKeyMapping:
    mapping: _PartitionsDefKeyMapping = defaultdict(set)

    entity_subsets_by_key = {es.key: es for es in entity_subsets}
    for key in entity_subsets_by_key:
        if isinstance(key, AssetCheckKey) and key.asset_key in entity_subsets_by_key:
            # all asset checks are currently unpartitioned, so ensure that they get grouped
            # into a run with the asset they check if that asset is being requested this
            # tick
            partitions_def = asset_graph.get(key.asset_key).partitions_def
            subset = entity_subsets_by_key[key.asset_key]
        else:
            partitions_def = asset_graph.get(key).partitions_def
            subset = entity_subsets_by_key[key]

        if partitions_def:
            for asset_partition in subset.expensively_compute_asset_partitions():
                mapping[partitions_def, asset_partition.partition_key].add(key)
        else:
            mapping[partitions_def, None].add(key)

    return mapping


def _build_backfill_request(
    entity_subsets: Sequence[EntitySubset[EntityKey]],
    asset_graph: BaseAssetGraph,
    run_tags: Optional[Mapping[str, str]],
) -> tuple[Optional[RunRequest], Sequence[EntitySubset[EntityKey]]]:
    """Determines a set of entity subsets that can be executed using a backfill.
    If any entity subset has size greater than 1, then it and all assets connected
    to it will be grouped into the backfill. Returns the corresponding backfill
    run request, and all entity subsets not handled in this process.
    """
    entity_subsets_by_key = {es.key: es for es in entity_subsets}
    visited: set[EntityKey] = set()
    backfill_subsets: list[EntitySubset[AssetKey]] = []

    def _flood_fill_asset_subsets(k: EntityKey):
        if k in visited or k not in entity_subsets_by_key:
            return []
        visited.add(k)
        if isinstance(k, AssetKey):
            node = asset_graph.get(k)
            if not node.is_materializable:
                # if the asset is not materializable, it cannot be included in a backfill
                return
            subset = cast(EntitySubset[AssetKey], entity_subsets_by_key.pop(k))
            backfill_subsets.append(subset)
            for sk in [
                # include all parent and child assets
                *node.parent_keys,
                *node.child_keys,
                # include all asset keys and check keys that must be executed alongside
                # this asset key
                *asset_graph.get_execution_set_asset_and_check_keys(k),
            ]:
                _flood_fill_asset_subsets(sk)
        else:
            # if we get an asset check in this code path, it must have been part of
            # the required execution set of an asset that is being backfilled, and
            # therefore the backfill daemon will handle executing it without being
            # explicitly told to, so just remove it from the set of things to consider
            entity_subsets_by_key.pop(k)

    # list() here because we modify the dict in flood_fill
    for k, es in list(entity_subsets_by_key.items()):
        if k in visited or es.size <= 1 or isinstance(k, AssetCheckKey):
            continue
        # this entity subset should be backfilled, as should all of its
        # attached assets
        _flood_fill_asset_subsets(k)

    backfill_request = (
        RunRequest.for_asset_graph_subset(
            asset_graph_subset=AssetGraphSubset.from_entity_subsets(backfill_subsets),
            tags=run_tags,
        )
        if backfill_subsets
        else None
    )
    return backfill_request, list(entity_subsets_by_key.values())


def build_run_requests(
    entity_subsets: Sequence[EntitySubset],
    asset_graph: BaseAssetGraph,
    run_tags: Optional[Mapping[str, str]],
    emit_backfills: bool,
) -> Sequence[RunRequest]:
    """For a single asset in a given tick, the asset will only be part of a run or a backfill, not both.
    If the asset is targetd by a backfill, there will only be one backfill that targets the asset.
    """
    if emit_backfills:
        backfill_run_request, entity_subsets = _build_backfill_request(
            entity_subsets, asset_graph, run_tags
        )
    else:
        backfill_run_request = None

    run_requests = _build_run_requests_from_partitions_def_mapping(
        _get_mapping_from_entity_subsets(entity_subsets, asset_graph),
        asset_graph,
        run_tags,
    )
    if backfill_run_request:
        run_requests = [backfill_run_request, *run_requests]

    return run_requests


def _build_run_requests_from_partitions_def_mapping(
    mapping: _PartitionsDefKeyMapping,
    asset_graph: BaseAssetGraph,
    run_tags: Optional[Mapping[str, str]],
) -> Sequence[RunRequest]:
    run_requests = []

    for (partitions_def, partition_key), entity_keys in mapping.items():
        tags = {**(run_tags or {})}
        if partition_key is not None:
            if partitions_def is None:
                check.failed("Partition key provided for unpartitioned entity")
            tags.update({**partitions_def.get_tags_for_partition_key(partition_key)})

        for entity_keys_for_repo in asset_graph.split_entity_keys_by_repository(entity_keys):
            asset_check_keys = [k for k in entity_keys_for_repo if isinstance(k, AssetCheckKey)]
            run_requests.append(
                # Do not call run_request.with_resolved_tags_and_config as the partition key is
                # valid and there is no config.
                # Calling with_resolved_tags_and_config is costly in asset reconciliation as it
                # checks for valid partition keys.
                RunRequest(
                    asset_selection=[k for k in entity_keys_for_repo if isinstance(k, AssetKey)],
                    partition_key=partition_key,
                    tags=tags,
                    # if selecting no asset_check_keys, just pass in `None` to allow required
                    # checks to be included
                    asset_check_keys=asset_check_keys or None,
                )
            )

    # We don't make public guarantees about sort order, but make an effort to provide a consistent
    # ordering that puts earlier time partitions before later time partitions. Note that, with dates
    # formatted like 12/7/2023, the run requests won't end up in time order. Sorting by converting
    # to time windows seemed more risky from a perf perspective, so we didn't include it here, but
    # it could make sense to actually benchmark that in the future.
    return sorted(run_requests, key=lambda x: x.partition_key if x.partition_key else "")


def build_run_requests_with_backfill_policies(
    asset_partitions: Iterable[AssetKeyPartitionKey],
    asset_graph: BaseAssetGraph,
    dynamic_partitions_store: DynamicPartitionsStore,
) -> Sequence[RunRequest]:
    """Build run requests for a selection of asset partitions based on the associated BackfillPolicies."""
    run_requests = []

    asset_partition_keys: Mapping[AssetKey, set[str]] = {
        asset_key_partition.asset_key: set() for asset_key_partition in asset_partitions
    }
    for asset_partition in asset_partitions:
        if asset_partition.partition_key:
            asset_partition_keys[asset_partition.asset_key].add(asset_partition.partition_key)

    assets_to_reconcile_by_partitions_def_partition_keys_backfill_policy: Mapping[
        tuple[Optional[PartitionsDefinition], Optional[frozenset[str]], Optional[BackfillPolicy]],
        set[AssetKey],
    ] = defaultdict(set)

    # here we are grouping assets by their partitions def, selected partition keys, and backfill policy.
    for asset_key, partition_keys in asset_partition_keys.items():
        assets_to_reconcile_by_partitions_def_partition_keys_backfill_policy[
            asset_graph.get(asset_key).partitions_def,
            frozenset(partition_keys) if partition_keys else None,
            asset_graph.get(asset_key).backfill_policy,
        ].add(asset_key)

    for (
        partitions_def,
        partition_keys,
        backfill_policy,
    ), asset_keys in assets_to_reconcile_by_partitions_def_partition_keys_backfill_policy.items():
        asset_check_keys = asset_graph.get_check_keys_for_assets(asset_keys)
        if partitions_def is None and partition_keys is not None:
            check.failed("Partition key provided for unpartitioned asset")
        elif partitions_def is not None and partition_keys is None:
            check.failed("Partition key missing for partitioned asset")
        elif partitions_def is None and partition_keys is None:
            # non partitioned assets will be backfilled in a single run
            run_requests.append(
                RunRequest(
                    asset_selection=list(asset_keys),
                    asset_check_keys=list(asset_check_keys),
                    tags={},
                )
            )
        elif backfill_policy is None:
            # just use the normal single-partition behavior
            entity_keys = cast(set[EntityKey], asset_keys)
            mapping: _PartitionsDefKeyMapping = {
                (partitions_def, pk): entity_keys for pk in (partition_keys or [None])
            }
            run_requests.extend(
                _build_run_requests_from_partitions_def_mapping(mapping, asset_graph, run_tags={})
            )
        else:
            run_requests.extend(
                _build_run_requests_with_backfill_policy(
                    list(asset_keys),
                    list(asset_check_keys),
                    check.not_none(backfill_policy),
                    check.not_none(partition_keys),
                    check.not_none(partitions_def),
                    tags={},
                    dynamic_partitions_store=dynamic_partitions_store,
                )
            )
    return run_requests


def _build_run_requests_with_backfill_policy(
    asset_keys: Sequence[AssetKey],
    asset_check_keys: Sequence[AssetCheckKey],
    backfill_policy: BackfillPolicy,
    partition_keys: frozenset[str],
    partitions_def: PartitionsDefinition,
    tags: dict[str, Any],
    dynamic_partitions_store: DynamicPartitionsStore,
) -> Sequence[RunRequest]:
    run_requests = []
    partition_subset = partitions_def.subset_with_partition_keys(partition_keys)
    partition_key_ranges = partition_subset.get_partition_key_ranges(
        partitions_def, dynamic_partitions_store=dynamic_partitions_store
    )
    for partition_key_range in partition_key_ranges:
        # We might resolve more than one partition key range for the given partition keys.
        # We can only apply chunking on individual partition key ranges.
        if backfill_policy.policy_type == BackfillPolicyType.SINGLE_RUN:
            run_requests.append(
                _build_run_request_for_partition_key_range(
                    asset_keys=list(asset_keys),
                    asset_check_keys=list(asset_check_keys),
                    partition_range_start=partition_key_range.start,
                    partition_range_end=partition_key_range.end,
                    run_tags=tags,
                )
            )
        else:
            run_requests.extend(
                _build_run_requests_for_partition_key_range(
                    asset_keys=list(asset_keys),
                    asset_check_keys=list(asset_check_keys),
                    partitions_def=partitions_def,
                    partition_key_range=partition_key_range,
                    max_partitions_per_run=check.int_param(
                        backfill_policy.max_partitions_per_run, "max_partitions_per_run"
                    ),
                    run_tags=tags,
                    dynamic_partitions_store=dynamic_partitions_store,
                )
            )
    return run_requests


def _build_run_requests_for_partition_key_range(
    asset_keys: Sequence[AssetKey],
    asset_check_keys: Sequence[AssetCheckKey],
    partitions_def: PartitionsDefinition,
    partition_key_range: PartitionKeyRange,
    max_partitions_per_run: int,
    run_tags: dict[str, str],
    dynamic_partitions_store: DynamicPartitionsStore,
) -> Sequence[RunRequest]:
    """Builds multiple run requests for the given partition key range. Each run request will have at most
    max_partitions_per_run partitions.
    """
    partition_keys = partitions_def.get_partition_keys_in_range(
        partition_key_range, dynamic_partitions_store=dynamic_partitions_store
    )
    partition_range_start_index = partition_keys.index(partition_key_range.start)
    partition_range_end_index = partition_keys.index(partition_key_range.end)

    partition_chunk_start_index = partition_range_start_index
    run_requests = []
    while partition_chunk_start_index <= partition_range_end_index:
        partition_chunk_end_index = partition_chunk_start_index + max_partitions_per_run - 1
        if partition_chunk_end_index > partition_range_end_index:
            partition_chunk_end_index = partition_range_end_index
        partition_chunk_start_key = partition_keys[partition_chunk_start_index]
        partition_chunk_end_key = partition_keys[partition_chunk_end_index]
        run_requests.append(
            _build_run_request_for_partition_key_range(
                asset_keys,
                asset_check_keys,
                partition_chunk_start_key,
                partition_chunk_end_key,
                run_tags,
            )
        )
        partition_chunk_start_index = partition_chunk_end_index + 1
    return run_requests


def _build_run_request_for_partition_key_range(
    asset_keys: Sequence[AssetKey],
    asset_check_keys: Sequence[AssetCheckKey],
    partition_range_start: str,
    partition_range_end: str,
    run_tags: dict[str, str],
) -> RunRequest:
    """Builds a single run request for the given asset key and partition key range."""
    tags = {
        **(run_tags or {}),
        ASSET_PARTITION_RANGE_START_TAG: partition_range_start,
        ASSET_PARTITION_RANGE_END_TAG: partition_range_end,
    }
    return RunRequest(
        asset_selection=asset_keys,
        partition_key=None,
        tags=tags,
        asset_check_keys=asset_check_keys,
    )



from typing import NamedTuple

from dagster._serdes import whitelist_for_serdes


# TimestampWithTimezone is used to preserve IANA timezone information when serializing.
# Serializing with the UTC offset (i.e. via datetime.isoformat) is insufficient because offsets vary
# depending on daylight savings time. This causes timedelta operations to be inexact, since the
# exact timezone is not preserved. To prevent any lossy serialization, ths implementation
# serializes both the datetime float and the IANA timezone.
@whitelist_for_serdes
class TimestampWithTimezone(NamedTuple):
    timestamp: float  # Seconds since the Unix epoch
    timezone: str



from dagster._core.container_context.config import (
    SHARED_CONTAINER_CONTEXT_SCHEMA as SHARED_CONTAINER_CONTEXT_SCHEMA,
    process_shared_container_context_config as process_shared_container_context_config,
)



import importlib
import importlib.util
import sys
from collections.abc import Mapping, Sequence
from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec, PathFinder
from types import ModuleType
from typing import Optional, Union


# The AliasedModuleFinder should be inserted in front of the built-in PathFinder.
def get_meta_path_insertion_index() -> int:
    for i in range(len(sys.meta_path)):
        finder = sys.meta_path[i]
        if isinstance(finder, type) and issubclass(finder, PathFinder):
            return i
    raise Exception(
        "Could not find the built-in PathFinder in sys.meta_path-- cannot insert the"
        " AliasedModuleFinder"
    )


class AliasedModuleFinder(MetaPathFinder):
    def __init__(self, alias_map: Mapping[str, str]):
        self.alias_map = alias_map

    def find_spec(
        self,
        fullname: str,
        _path: Optional[Sequence[Union[bytes, str]]] = None,
        _target: Optional[ModuleType] = None,
    ) -> Optional[ModuleSpec]:
        head = next((k for k in self.alias_map.keys() if fullname.startswith(k)), None)
        if head is not None:
            base_name = self.alias_map[head] + fullname[len(head) :]
            base_spec = importlib.util.find_spec(base_name)
            assert base_spec, f"Could not find module spec for {base_name}."
            return ModuleSpec(
                fullname,
                AliasedModuleLoader(fullname, base_spec),
                origin=base_spec.origin,
                is_package=base_spec.submodule_search_locations is not None,
            )
        else:
            return None


# Key reference to understand the load process:
#   https://docs.python.org/3/reference/import.html#loading


# While it is possible to override `Loader.create_module` to simply return the base module, this
# is undesirable because the import system modifies the module's metadata attributes after
# creation and outside of our control. This means that, if we simply had:
#
#     def create_module(self, spec):
#         return importlib.import_module(self.base_spec_name)
#
# The returned base module would have its name etc modified, e.g. the already-loaded
# `dagster._core` would be renamed to alias `dagster.core`. To avoid this, we let the import system
# generate a module using default logic, then simply discard this module in `exec_module` (the final
# step), where it is passed in. This is the point at which we swap in the base module, which we
# obtain through `importlib.import_module` (like a standard import statetment, this will simply
# returned the cached module from `sys.modules` if it has already been loaded). The swap is done by
# simply replacing the dummy module (already stored in `sys.modules` outside of our control) with
# the imported base module.
class AliasedModuleLoader(Loader):
    def __init__(self, alias: str, base_spec: ModuleSpec):
        self.alias = alias
        self.base_spec = base_spec

    def exec_module(self, _module: ModuleType) -> None:
        base_module = importlib.import_module(self.base_spec.name)
        sys.modules[self.alias] = base_module

    def module_repr(self, module: ModuleType) -> str:
        assert self.base_spec.loader
        return self.base_spec.loader.module_repr(module)



"""Internal py2/3 compatibility library. A little more than six."""

import inspect
import os
import shlex
import sys
import time
from collections.abc import Sequence
from types import ModuleType
from typing import Any, Callable, List, Type  # noqa: F401, UP035

from typing_extensions import TypeGuard

import dagster._seven.json as json  # noqa: F401
from dagster._seven.json import (
    JSONDecodeError as JSONDecodeError,
    dump as dump,
    dumps as dumps,
)
from dagster._seven.temp_dir import get_system_temp_directory as get_system_temp_directory

IS_WINDOWS = os.name == "nt"
IS_PYTHON_3_12 = (sys.version_info[0], sys.version_info[1]) == (3, 12)

# TODO implement a generic import by name -- see https://stackoverflow.com/questions/301134/how-to-import-a-module-given-its-name


# https://stackoverflow.com/a/67692/324449
def import_module_from_path(module_name: str, path_to_file: str) -> ModuleType:
    import importlib.util

    spec = importlib.util.spec_from_file_location(module_name, path_to_file)
    if spec is None:
        raise Exception(
            f"Can not import module {module_name} from path {path_to_file}, unable to load spec."
        )

    if sys.modules.get(spec.name) and spec.origin:
        f = sys.modules[spec.name].__file__
        # __file__ can be relative depending on current working directory
        if f and os.path.abspath(f) == os.path.abspath(spec.origin):
            return sys.modules[spec.name]

    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    assert spec.loader
    spec.loader.exec_module(module)
    return module


def import_uncached_module_from_path(module_name: str, path_to_file: str) -> ModuleType:
    import importlib.util

    spec = importlib.util.spec_from_file_location(module_name, path_to_file)
    if spec is None:
        raise Exception(
            f"Can not import module {module_name} from path {path_to_file}, unable to load spec."
        )

    module = importlib.util.module_from_spec(spec)
    assert spec.loader
    spec.loader.exec_module(module)
    return module


time_fn = time.perf_counter


def get_arg_names(callable_: Callable[..., Any]) -> Sequence[str]:
    return [
        parameter.name
        for parameter in inspect.signature(callable_).parameters.values()
        if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
    ]


# https://stackoverflow.com/a/58437485/324449
def is_module_available(module_name: str) -> bool:
    # python 3.4 and above
    import importlib.util

    loader = importlib.util.find_spec(module_name)

    return loader is not None


def is_lambda(target: object) -> TypeGuard[Callable[..., Any]]:
    return callable(target) and getattr(target, "__name__", None) == "<lambda>"


def is_function_or_decorator_instance_of(
    target: object, kls: type[Any]
) -> TypeGuard[Callable[..., Any]]:
    return inspect.isfunction(target) or (isinstance(target, kls) and hasattr(target, "__name__"))


def qualname_differs(target: object) -> bool:
    return hasattr(target, "__qualname__") and getattr(target, "__qualname__") != getattr(
        target, "__name__"
    )


def xplat_shlex_split(s: str) -> list[str]:
    if IS_WINDOWS:
        return shlex.split(s, posix=False)

    return shlex.split(s)


def get_import_error_message(import_error: ImportError) -> str:
    return import_error.msg


def is_subclass(child_type: type[Any], parent_type: type[Any]):
    """Due to some pathological interactions betwen bugs in the Python typing library
    (https://github.com/python/cpython/issues/88459 and
    https://github.com/python/cpython/issues/89010), some types (list[str] in Python 3.9, for
    example) pass inspect.isclass check above but then raise an exception if issubclass is called
    with the same class. This function provides a workaround for that issue.
    """
    if not inspect.isclass(child_type):
        return False

    try:
        return issubclass(child_type, parent_type)
    except TypeError:
        return False



from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Callable, Optional, Union

from typing_extensions import TypeAlias

from dagster import _check as check
from dagster._config import UserConfigSchema
from dagster._core.decorator_utils import format_docstring_for_description
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import AssetCheckSpec
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.asset_dep import AssetDep, CoercibleToAssetDep
from dagster._core.definitions.asset_in import AssetIn, CoercibleToAssetIn
from dagster._core.definitions.asset_key import AssetCheckKey
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.decorators.asset_decorator import make_asset_deps
from dagster._core.definitions.decorators.decorator_assets_definition_builder import (
    DecoratorAssetsDefinitionBuilder,
    DecoratorAssetsDefinitionBuilderArgs,
    NamedIn,
    build_and_validate_named_ins,
    compute_required_resource_keys,
    get_function_params_without_context_or_config_or_resources,
    validate_named_ins_subset_of_deps,
)
from dagster._core.definitions.decorators.op_decorator import _Op
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.output import Out
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.utils import DEFAULT_OUTPUT
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.execution.build_resources import wrap_resources_for_execution
from dagster._core.storage.tags import COMPUTE_KIND_TAG
from dagster._utils.warnings import disable_dagster_warnings

AssetCheckFunctionReturn: TypeAlias = AssetCheckResult
AssetCheckFunction: TypeAlias = Callable[..., AssetCheckFunctionReturn]


def _build_asset_check_named_ins(
    name: str,
    asset_key: AssetKey,
    fn: Callable[..., Any],
    additional_ins: Mapping[str, AssetIn],
    additional_deps: Mapping[AssetKey, AssetDep],
) -> Mapping[AssetKey, NamedIn]:
    fn_params = get_function_params_without_context_or_config_or_resources(fn)

    if asset_key in (additional_deps or []):
        raise DagsterInvalidDefinitionError(
            f"When defining check '{name}', asset '{asset_key.to_user_string()}' was passed to `asset` and `additional_deps`."
            " It can only be passed to one of these parameters."
        )
    if asset_key in [asset_in.key for asset_in in additional_ins.values()]:
        raise DagsterInvalidDefinitionError(
            f"When defining check '{name}', asset '{asset_key.to_user_string()}' was passed to `asset` and `additional_ins`."
            " It can only be passed to one of these parameters."
        )

    fn_param_names = {param.name for param in fn_params}
    for in_name in additional_ins.keys():
        if in_name not in fn_param_names:
            raise DagsterInvalidDefinitionError(
                f"'{in_name}' is specified in 'additional_ins' but isn't a parameter."
            )

    # if all the fn_params are in additional_ins, then we add the primary asset as a dep
    if len(fn_params) == len(additional_ins):
        all_deps = {**additional_deps, **{asset_key: AssetDep(asset_key)}}
        all_ins = additional_ins
    # otherwise there should be one extra fn_param, which is the primary asset. Add that as an input
    elif len(fn_params) == len(additional_ins) + 1:
        primary_asset_param_name = next(
            param.name for param in fn_params if param.name not in additional_ins.keys()
        )
        all_ins = {**additional_ins, primary_asset_param_name: AssetIn(asset_key)}
        all_deps = additional_deps
    else:
        param_names_not_in_additional_ins = sorted(
            [f"'{name}'" for name in (fn_param_names - set(additional_ins.keys()))]
        )
        raise DagsterInvalidDefinitionError(
            f"When defining check '{name}', multiple assets provided as parameters:"
            f" [{', '.join(param_names_not_in_additional_ins)}]. These should either match"
            " the target asset or be specified in 'additional_ins'."
        )

    return build_and_validate_named_ins(
        fn=fn,
        asset_ins=all_ins,
        deps=all_deps.values(),
    )


def asset_check(
    *,
    asset: Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset],
    name: Optional[str] = None,
    description: Optional[str] = None,
    blocking: bool = False,
    additional_ins: Optional[Mapping[str, AssetIn]] = None,
    additional_deps: Optional[Iterable[CoercibleToAssetDep]] = None,
    required_resource_keys: Optional[set[str]] = None,
    resource_defs: Optional[Mapping[str, object]] = None,
    config_schema: Optional[UserConfigSchema] = None,
    compute_kind: Optional[str] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    retry_policy: Optional[RetryPolicy] = None,
    metadata: Optional[Mapping[str, Any]] = None,
    automation_condition: Optional[AutomationCondition[AssetCheckKey]] = None,
) -> Callable[[AssetCheckFunction], AssetChecksDefinition]:
    """Create a definition for how to execute an asset check.

    Args:
        asset (Union[AssetKey, Sequence[str], str, AssetsDefinition, SourceAsset]): The
            asset that the check applies to.
        name (Optional[str]): The name of the check. If not specified, the name of the decorated
            function will be used. Checks for the same asset must have unique names.
        description (Optional[str]): The description of the check.
        blocking (bool): When enabled, runs that include this check and any downstream assets that
            depend on `asset` will wait for this check to complete before starting the downstream
            assets. If the check fails with severity `AssetCheckSeverity.ERROR`, then the downstream
            assets won't execute.
        additional_ins (Optional[Mapping[str, AssetIn]]): A mapping from input name to
            information about the input. These inputs will apply to the underlying op that
            executes the check. These should not include the `asset` parameter, which is
            always included as a dependency.
        additional_deps (Optional[Iterable[CoercibleToAssetDep]]): Assets that are upstream
            dependencies, but do not correspond to a parameter of the decorated function. These
            dependencies will apply to the underlying op that executes the check. These should not
            include the `asset` parameter, which is always included as a dependency.
        required_resource_keys (Optional[Set[str]]): A set of keys for resources that are required
            by the function that execute the check. These can alternatively be specified by
            including resource-typed parameters in the function signature.
        config_schema (Optional[ConfigSchema): The configuration schema for the check's underlying
            op. If set, Dagster will check that config provided for the op matches this schema and fail
            if it does not. If not set, Dagster will accept any config provided for the op.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that executes the check.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        compute_kind (Optional[str]): A string to represent the kind of computation that executes
            the check, e.g. "dbt" or "spark".
        retry_policy (Optional[RetryPolicy]): The retry policy for the op that executes the check.
        metadata (Optional[Mapping[str, Any]]): A dictionary of static metadata for the check.
        automation_condition (Optional[AutomationCondition]): An AutomationCondition which determines
            when this check should be executed.

    Produces an :py:class:`AssetChecksDefinition` object.


    Example:
        .. code-block:: python

            from dagster import asset, asset_check, AssetCheckResult

            @asset
            def my_asset() -> None:
                ...

            @asset_check(asset=my_asset, description="Check that my asset has enough rows")
            def my_asset_has_enough_rows() -> AssetCheckResult:
                num_rows = ...
                return AssetCheckResult(passed=num_rows > 5, metadata={"num_rows": num_rows})


    Example with a DataFrame Output:
        .. code-block:: python

            from dagster import asset, asset_check, AssetCheckResult
            from pandas import DataFrame

            @asset
            def my_asset() -> DataFrame:
                ...

            @asset_check(asset=my_asset, description="Check that my asset has enough rows")
            def my_asset_has_enough_rows(my_asset: DataFrame) -> AssetCheckResult:
                num_rows = my_asset.shape[0]
                return AssetCheckResult(passed=num_rows > 5, metadata={"num_rows": num_rows})
    """
    check.opt_set_param(required_resource_keys, "required_resource_keys", of_type=str)
    dict(check.opt_mapping_param(resource_defs, "resource_defs", key_type=str))

    def inner(fn: AssetCheckFunction) -> AssetChecksDefinition:
        check.callable_param(fn, "fn")
        resolved_name = name or fn.__name__
        asset_key = AssetKey.from_coercible_or_definition(asset)

        additional_dep_keys = (
            {dep.asset_key: dep for dep in make_asset_deps(additional_deps) or []}
            if additional_deps
            else {}
        )
        named_in_by_asset_key = _build_asset_check_named_ins(
            resolved_name,
            asset_key,
            fn,
            additional_ins=additional_ins or {},
            additional_deps=additional_dep_keys,
        )

        # additional_deps on AssetCheckSpec holds the keys passed to additional_deps and
        # additional_ins. We don't want to include the primary asset key in this set.
        additional_ins_and_deps = named_in_by_asset_key.keys() - {asset_key}

        spec = AssetCheckSpec(
            name=resolved_name,
            description=description or format_docstring_for_description(fn),
            asset=asset_key,
            additional_deps=additional_ins_and_deps,
            blocking=blocking,
            metadata=metadata,
            automation_condition=automation_condition,
        )

        resource_defs_for_execution = wrap_resources_for_execution(resource_defs)

        builder_args = DecoratorAssetsDefinitionBuilderArgs(
            decorator_name="@asset_check",
            name=name,
            # @asset_check previous behavior is to not set description on underlying op
            op_description=None,
            required_resource_keys=required_resource_keys or set(),
            config_schema=config_schema,
            retry_policy=retry_policy,
            specs=[],
            check_specs_by_output_name={DEFAULT_OUTPUT: spec},
            can_subset=False,
            compute_kind=compute_kind,
            op_tags=op_tags,
            op_def_resource_defs=resource_defs_for_execution,
            assets_def_resource_defs=resource_defs_for_execution,
            upstream_asset_deps=[],
            # unsupported capabiltiies in asset checks
            partitions_def=None,
            code_version=None,
            backfill_policy=None,
            group_name=None,
            # non-sensical args in this context
            asset_deps={},
            asset_in_map={},
            asset_out_map={},
            execution_type=None,
            pool=None,
        )

        builder = DecoratorAssetsDefinitionBuilder(
            named_ins_by_asset_key=named_in_by_asset_key,
            named_outs_by_asset_key={},
            internal_deps={},
            op_name=spec.get_python_identifier(),
            args=builder_args,
            fn=fn,
        )

        op_def = builder.create_op_definition()

        return AssetChecksDefinition.create(
            keys_by_input_name={
                input_tuple[0]: asset_key
                for asset_key, input_tuple in named_in_by_asset_key.items()
            },
            node_def=op_def,
            resource_defs=resource_defs_for_execution,
            check_specs_by_output_name={op_def.output_defs[0].name: spec},
            can_subset=False,
        )

    return inner


MultiAssetCheckFunctionReturn: TypeAlias = Iterable[AssetCheckResult]
MultiAssetCheckFunction: TypeAlias = Callable[..., MultiAssetCheckFunctionReturn]


def multi_asset_check(
    *,
    name: Optional[str] = None,
    specs: Sequence[AssetCheckSpec],
    description: Optional[str] = None,
    can_subset: bool = False,
    compute_kind: Optional[str] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    resource_defs: Optional[Mapping[str, object]] = None,
    required_resource_keys: Optional[set[str]] = None,
    retry_policy: Optional[RetryPolicy] = None,
    config_schema: Optional[UserConfigSchema] = None,
    ins: Optional[Mapping[str, CoercibleToAssetIn]] = None,
) -> Callable[[Callable[..., Any]], AssetChecksDefinition]:
    """Defines a set of asset checks that can be executed together with the same op.

    Args:
        specs (Sequence[AssetCheckSpec]): Specs for the asset checks.
        name (Optional[str]): The name of the op. If not specified, the name of the decorated
            function will be used.
        description (Optional[str]): Description of the op.
        required_resource_keys (Optional[Set[str]]): A set of keys for resources that are required
            by the function that execute the checks. These can alternatively be specified by
            including resource-typed parameters in the function signature.
        config_schema (Optional[ConfigSchema): The configuration schema for the asset checks' underlying
            op. If set, Dagster will check that config provided for the op matches this schema and fail
            if it does not. If not set, Dagster will accept any config provided for the op.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that executes the checks.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        compute_kind (Optional[str]): A string to represent the kind of computation that executes
            the checks, e.g. "dbt" or "spark".
        retry_policy (Optional[RetryPolicy]): The retry policy for the op that executes the checks.
        can_subset (bool): Whether the op can emit results for a subset of the asset checks
            keys, based on the context.selected_asset_check_keys argument. Defaults to False.
        ins (Optional[Mapping[str, Union[AssetKey, AssetIn]]]): A mapping from input name to AssetIn depended upon by
            a given asset check. If an AssetKey is provided, it will be converted to an AssetIn with the same key.


    Examples:
        .. code-block:: python

            @multi_asset_check(
                specs=[
                    AssetCheckSpec("enough_rows", asset="asset1"),
                    AssetCheckSpec("no_dupes", asset="asset1"),
                    AssetCheckSpec("enough_rows", asset="asset2"),
                ],
            )
            def checks():
                yield AssetCheckResult(passed=True, asset_key="asset1", check_name="enough_rows")
                yield AssetCheckResult(passed=False, asset_key="asset1", check_name="no_dupes")
                yield AssetCheckResult(passed=True, asset_key="asset2", check_name="enough_rows")

    """
    required_resource_keys = check.opt_set_param(
        required_resource_keys, "required_resource_keys", of_type=str
    )
    resource_defs = wrap_resources_for_execution(
        check.opt_mapping_param(resource_defs, "resource_defs", key_type=str)
    )
    config_schema = check.opt_mapping_param(
        config_schema,  # type: ignore
        "config_schema",
        additional_message="Only dicts are supported for asset config_schema.",
    )

    def inner(fn: MultiAssetCheckFunction) -> AssetChecksDefinition:
        op_name = name or fn.__name__
        op_required_resource_keys = compute_required_resource_keys(
            required_resource_keys, resource_defs, fn=fn, decorator_name="@multi_asset_check"
        )

        outs = {
            spec.get_python_identifier(): Out(None, is_required=not can_subset) for spec in specs
        }
        all_deps_by_key = {
            **{spec.asset_key: AssetDep(spec.asset_key) for spec in specs},
            **{dep.asset_key: dep for spec in specs for dep in (spec.additional_deps or [])},
        }

        named_ins_by_asset_key = build_and_validate_named_ins(
            fn=fn,
            asset_ins={
                inp_name: AssetIn.from_coercible(coercible) for inp_name, coercible in ins.items()
            }
            if ins
            else {},
            deps=all_deps_by_key.values(),
        )
        validate_named_ins_subset_of_deps(named_ins_by_asset_key, all_deps_by_key)

        with disable_dagster_warnings():
            op_def = _Op(
                name=op_name,
                description=description,
                ins=dict(named_ins_by_asset_key.values()),
                out=outs,
                required_resource_keys=op_required_resource_keys,
                tags={
                    **({COMPUTE_KIND_TAG: compute_kind} if compute_kind else {}),
                    **(op_tags or {}),
                },
                config_schema=config_schema,
                retry_policy=retry_policy,
            )(fn)

        return AssetChecksDefinition.create(
            node_def=op_def,
            resource_defs=wrap_resources_for_execution(resource_defs),
            keys_by_input_name={
                input_tuple[0]: asset_key
                for asset_key, input_tuple in named_ins_by_asset_key.items()
            },
            check_specs_by_output_name={spec.get_python_identifier(): spec for spec in specs},
            can_subset=can_subset,
        )

    return inner



from typing import Optional

import click

import dagster._check as check
from dagster._cli.job import get_run_config_from_cli_opts
from dagster._cli.utils import (
    assert_no_remaining_opts,
    get_instance_for_cli,
    get_possibly_temporary_instance_for_cli,
)
from dagster._cli.workspace.cli_target import (
    PythonPointerOpts,
    get_repository_python_origin_from_cli_opts,
    python_pointer_options,
    run_config_option,
)
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.backfill_policy import BackfillPolicyType
from dagster._core.definitions.events import AssetKey
from dagster._core.errors import DagsterInvalidSubsetError, DagsterUnknownPartitionError
from dagster._core.execution.api import execute_job
from dagster._core.instance import DagsterInstance
from dagster._core.origin import JobPythonOrigin
from dagster._core.storage.tags import (
    ASSET_PARTITION_RANGE_END_TAG,
    ASSET_PARTITION_RANGE_START_TAG,
)
from dagster._core.telemetry import telemetry_wrapper
from dagster._utils.hosted_user_process import recon_job_from_origin, recon_repository_from_origin
from dagster._utils.interrupts import capture_interrupts


@click.group(name="asset")
def asset_cli():
    """Commands for working with Dagster assets."""


@asset_cli.command(name="materialize", help="Execute a run to materialize a selection of assets")
@click.option("--select", help="Comma-separated Asset selection to target", required=True)
@click.option("--partition", help="Asset partition to target", required=False)
@click.option(
    "--partition-range",
    help="Asset partition range to target i.e. <start>...<end>",
    required=False,
)
@click.option(
    "--config-json",
    type=click.STRING,
    help="JSON string of run config to use for this job run. Cannot be used with -c / --config.",
)
@run_config_option(name="config", command_name="materialize")
@python_pointer_options
def asset_materialize_command(
    select: str,
    partition: Optional[str],
    partition_range: Optional[str],
    config: tuple[str, ...],
    config_json: Optional[str],
    **other_opts: object,
) -> None:
    python_pointer_opts = PythonPointerOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)

    with capture_interrupts():
        with get_possibly_temporary_instance_for_cli(
            "``dagster asset materialize``",
        ) as instance:
            execute_materialize_command(
                instance=instance,
                select=select,
                partition=partition,
                partition_range=partition_range,
                config=config,
                config_json=config_json,
                python_pointer_opts=python_pointer_opts,
            )


@telemetry_wrapper
def execute_materialize_command(
    *,
    instance: DagsterInstance,
    select: str,
    partition: Optional[str],
    partition_range: Optional[str],
    config: tuple[str, ...],
    config_json: Optional[str],
    python_pointer_opts: PythonPointerOpts,
) -> None:
    run_config = get_run_config_from_cli_opts(config, config_json)
    repository_origin = get_repository_python_origin_from_cli_opts(python_pointer_opts)

    recon_repo = recon_repository_from_origin(repository_origin)
    repo_def = recon_repo.get_definition()

    asset_selection = AssetSelection.from_coercible(select.split(","))
    asset_keys = asset_selection.resolve(repo_def.asset_graph)

    implicit_job_def = repo_def.get_implicit_job_def_for_assets(asset_keys)
    # If we can't find an implicit job with all the given assets, it's because they couldn't be
    # placed into the same implicit job, because of their conflicting PartitionsDefinitions.
    if implicit_job_def is None:
        raise DagsterInvalidSubsetError(
            "All selected assets must share the same PartitionsDefinition or have no"
            " PartitionsDefinition"
        )

    reconstructable_job = recon_job_from_origin(
        JobPythonOrigin(implicit_job_def.name, repository_origin=repository_origin)
    )

    if partition and partition_range:
        check.failed("Cannot specify both --partition and --partition-range options. Use only one.")

    if partition:
        if all(
            implicit_job_def.asset_layer.get(asset_key).partitions_def is None
            for asset_key in asset_keys
        ):
            check.failed("Provided '--partition' option, but none of the assets are partitioned")

        try:
            implicit_job_def.validate_partition_key(
                partition, selected_asset_keys=asset_keys, dynamic_partitions_store=instance
            )
            tags = implicit_job_def.get_tags_for_partition_key(
                partition, selected_asset_keys=asset_keys
            )
        except DagsterUnknownPartitionError:
            raise DagsterInvalidSubsetError(
                "All selected assets must have a PartitionsDefinition containing the passed"
                f" partition key `{partition}` or have no PartitionsDefinition."
            )
    elif partition_range:
        if len(partition_range.split("...")) != 2:
            check.failed("Invalid partition range format. Expected <start>...<end>.")

        partition_range_start, partition_range_end = partition_range.split("...")

        for asset_key in asset_keys:
            backfill_policy = implicit_job_def.asset_layer.get(asset_key).backfill_policy
            if (
                backfill_policy is not None
                and backfill_policy.policy_type != BackfillPolicyType.SINGLE_RUN
            ):
                check.failed(
                    "Provided partition range, but not all assets have a single-run backfill policy."
                )
        try:
            implicit_job_def.validate_partition_key(
                partition_range_start,
                selected_asset_keys=asset_keys,
                dynamic_partitions_store=instance,
            )
            implicit_job_def.validate_partition_key(
                check.not_none(partition_range_end),
                selected_asset_keys=asset_keys,
                dynamic_partitions_store=instance,
            )
        except DagsterUnknownPartitionError:
            raise DagsterInvalidSubsetError(
                "All selected assets must have a PartitionsDefinition containing the passed"
                f" partition key `{partition_range_start}` or have no PartitionsDefinition."
            )
        tags = {
            ASSET_PARTITION_RANGE_START_TAG: partition_range_start,
            ASSET_PARTITION_RANGE_END_TAG: partition_range_end,
        }
    else:
        if any(
            implicit_job_def.asset_layer.get(asset_key).partitions_def is not None
            for asset_key in asset_keys
        ):
            check.failed("Asset has partitions, but no '--partition' option was provided")
        tags = {}

    result = execute_job(
        job=reconstructable_job,
        asset_selection=list(asset_keys),
        instance=instance,
        tags=tags,
        run_config=run_config,
    )
    if not result.success:
        raise click.ClickException("Materialization failed.")


@asset_cli.command(name="list", help="List assets")
@click.option("--select", help="Asset selection to target", required=False)
@python_pointer_options
def asset_list_command(select: Optional[str], **other_opts):
    python_pointer_opts = PythonPointerOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)

    repository_origin = get_repository_python_origin_from_cli_opts(python_pointer_opts)
    recon_repo = recon_repository_from_origin(repository_origin)
    repo_def = recon_repo.get_definition()

    if select is not None:
        asset_selection = AssetSelection.from_coercible(select.split(","))
    else:
        asset_selection = AssetSelection.all()

    asset_keys = asset_selection.resolve(repo_def.asset_graph, allow_missing=True)

    for asset_key in sorted(asset_keys):
        print(asset_key.to_user_string())  # noqa: T201


@asset_cli.command(name="wipe")
@click.argument("key", nargs=-1)
@click.option("--all", is_flag=True, help="Eliminate all asset key indexes")
@click.option("--noprompt", is_flag=True)
def asset_wipe_command(key, **cli_args):
    r"""Eliminate asset key indexes from event logs.

    Warning: Cannot be undone.

    \b
    Usage:
      dagster asset wipe --all
      dagster asset wipe <unstructured_asset_key_name>
      dagster asset wipe <json_string_of_structured_asset_key>
    """
    if not cli_args.get("all") and len(key) == 0:
        raise click.UsageError(
            "Error, you must specify an asset key or use `--all` to wipe all asset keys."
        )

    if cli_args.get("all") and len(key) > 0:
        raise click.UsageError("Error, cannot use more than one of: asset key, `--all`.")

    noprompt = cli_args.get("noprompt")

    with get_instance_for_cli() as instance:
        if len(key) > 0:
            asset_keys = [AssetKey.from_db_string(key_string) for key_string in key]
            prompt = (
                "Are you sure you want to remove the asset key indexes for these keys from the"
                " event logs? Type DELETE"
            )
        else:
            asset_keys = instance.all_asset_keys()
            prompt = (
                "Are you sure you want to remove all asset key indexes from the event logs? Type"
                " DELETE"
            )

        if noprompt:
            confirmation = "DELETE"
        else:
            confirmation = click.prompt(prompt)

        if confirmation == "DELETE":
            instance.wipe_assets(asset_keys)  # pyright: ignore[reportArgumentType]
            click.echo("Removed asset indexes from event logs")
        else:
            click.echo("Exiting without removing asset indexes")


@asset_cli.command(name="wipe-partitions-status-cache")
@click.argument("key", nargs=-1)
@click.option("--all", is_flag=True, help="Wipe partitions status cache of all asset keys")
@click.option("--noprompt", is_flag=True)
def asset_wipe_cache_command(key, **cli_args):
    r"""Clears the asset partitions status cache, which is used by the webserver to load partition
    pages more quickly. The cache will be rebuilt the next time the partition pages are loaded,
    if caching is enabled.

    \b
    Usage:
      dagster asset wipe-cache --all
      dagster asset wipe-cache <unstructured_asset_key_name>
      dagster asset wipe-cache <json_string_of_structured_asset_key>
    """
    if not cli_args.get("all") and len(key) == 0:
        raise click.UsageError(
            "Error, you must specify an asset key or use `--all` to clear the partitions status"
            " cache of all asset keys."
        )

    if cli_args.get("all") and len(key) > 0:
        raise click.UsageError("Error, cannot use more than one of: asset key, `--all`.")

    noprompt = cli_args.get("noprompt")

    with get_instance_for_cli() as instance:
        if instance.can_read_asset_status_cache() is False:
            raise click.UsageError(
                "Error, the instance does not support caching asset status. Wiping the cache is not"
                " supported."
            )

        if len(key) > 0:
            asset_keys = [AssetKey.from_db_string(key_string) for key_string in key]
            prompt = (
                "Are you sure you want to wipe the partitions status cache for these"
                f" keys {asset_keys} from the event logs? Type DELETE"
            )
        else:
            asset_keys = instance.all_asset_keys()
            prompt = (
                "Are you sure you want to wipe the partitions status cache for all asset keys?"
                " Type DELETE"
            )

        if noprompt:
            confirmation = "DELETE"
        else:
            confirmation = click.prompt(prompt)

        if confirmation == "DELETE":
            instance.wipe_asset_cached_status(asset_keys)  # pyright: ignore[reportArgumentType]
            click.echo("Cleared the partitions status cache")
        else:
            click.echo("Exiting without wiping the partitions status cache")



import datetime
import json
import os
import sys
import tempfile
import time
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from threading import Event, Thread
from typing import IO, Optional, TypeVar, Union

from dagster_pipes import (
    PIPES_PROTOCOL_VERSION_FIELD,
    PipesContextData,
    PipesDefaultContextLoader,
    PipesDefaultMessageWriter,
    PipesExtras,
    PipesOpenedData,
    PipesParams,
)

from dagster import (
    OpExecutionContext,
    _check as check,
)
from dagster._core.errors import DagsterInvariantViolationError, DagsterPipesExecutionError
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.pipes.client import PipesContextInjector, PipesLaunchedData, PipesMessageReader
from dagster._core.pipes.context import (
    PipesMessageHandler,
    PipesSession,
    build_external_execution_context_data,
)
from dagster._utils import tail_file

TCursor = TypeVar("TCursor")


_CONTEXT_INJECTOR_FILENAME = "context"
_MESSAGE_READER_FILENAME = "messages"


class PipesFileContextInjector(PipesContextInjector):
    """Context injector that injects context data into the external process by writing it to a
    specified file.

    Args:
        path (str): The path of a file to which to write context data. The file will be deleted on
            close of the pipes session.
    """

    def __init__(self, path: str):
        self._path = check.str_param(path, "path")

    @contextmanager
    def inject_context(self, context_data: "PipesContextData") -> Iterator[PipesParams]:
        """Inject context to external environment by writing it to a file as JSON and exposing the
        path to the file.

        Args:
            context_data (PipesContextData): The context data to inject.

        Yields:
            PipesParams: A dict of parameters that can be used by the external process to locate and
            load the injected context data.
        """
        with open(self._path, "w") as input_stream:
            json.dump(context_data, input_stream)
        try:
            yield {PipesDefaultContextLoader.FILE_PATH_KEY: self._path}
        finally:
            if os.path.exists(self._path):
                os.remove(self._path)

    def no_messages_debug_text(self) -> str:
        return f"Attempted to inject context via file {self._path}"


class PipesTempFileContextInjector(PipesContextInjector):
    """Context injector that injects context data into the external process by writing it to an
    automatically-generated temporary file.
    """

    @contextmanager
    def inject_context(self, context: "PipesContextData") -> Iterator[PipesParams]:
        """Inject context to external environment by writing it to an automatically-generated
        temporary file as JSON and exposing the path to the file.

        Args:
            context_data (PipesContextData): The context data to inject.

        Yields:
            PipesParams: A dict of parameters that can be used by the external process to locate and
            load the injected context data.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            with PipesFileContextInjector(
                os.path.join(tempdir, _CONTEXT_INJECTOR_FILENAME)
            ).inject_context(context) as params:
                yield params

    def no_messages_debug_text(self) -> str:
        return "Attempted to inject context via a temporary file."


class PipesEnvContextInjector(PipesContextInjector):
    """Context injector that injects context data into the external process by injecting it directly into the external process environment."""

    @contextmanager
    def inject_context(
        self,
        context_data: "PipesContextData",
    ) -> Iterator[PipesParams]:
        """Inject context to external environment by embedding directly in the parameters that will
        be passed to the external process (typically as environment variables).

        Args:
            context_data (PipesContextData): The context data to inject.

        Yields:
            PipesParams: A dict of parameters that can be used by the external process to locate and
            load the injected context data.
        """
        yield {PipesDefaultContextLoader.DIRECT_KEY: context_data}

    def no_messages_debug_text(self) -> str:
        return "Attempted to inject context directly, typically as an environment variable."


class PipesFileMessageReader(PipesMessageReader):
    """Message reader that reads messages by tailing a specified file.

    Args:
        path (str): The path of the file to which messages will be written. The file will be deleted
            on close of the pipes session.
        include_stdio_in_messages (bool): Whether to include stdout/stderr logs in the messages produced by the message writer in the external process.
        cleanup_file (bool): Whether to delete the file on close of the pipes session.
    """

    def __init__(
        self, path: str, include_stdio_in_messages: bool = False, cleanup_file: bool = True
    ):
        self._path = check.str_param(path, "path")
        self._include_stdio_in_messages = check.bool_param(
            include_stdio_in_messages, "include_stdio_in_messages"
        )
        self._cleanup_file = cleanup_file

    def on_launched(self, params: PipesLaunchedData) -> None:
        self.launched_payload = params

    @contextmanager
    def read_messages(
        self,
        handler: "PipesMessageHandler",
    ) -> Iterator[PipesParams]:
        """Set up a thread to read streaming messages from the external process by tailing the
        target file.

        Args:
            handler (PipesMessageHandler): object to process incoming messages

        Yields:
            PipesParams: A dict of parameters that specifies where a pipes process should write
            pipes protocol messages.
        """
        is_session_closed = Event()
        thread = None
        try:
            open(self._path, "w").close()  # create file
            thread = Thread(
                target=self._reader_thread,
                args=(handler, is_session_closed),
                daemon=True,
            )
            thread.start()
            yield {
                PipesDefaultMessageWriter.FILE_PATH_KEY: self._path,
                PipesDefaultMessageWriter.INCLUDE_STDIO_IN_MESSAGES_KEY: self._include_stdio_in_messages,
            }
        finally:
            is_session_closed.set()
            if thread:
                thread.join()
            if os.path.exists(self._path) and self._cleanup_file:
                os.remove(self._path)

    def _reader_thread(self, handler: "PipesMessageHandler", is_resource_complete: Event) -> None:
        try:
            for line in tail_file(self._path, lambda: is_resource_complete.is_set()):
                message = json.loads(line)
                handler.handle_message(message)
        except:
            handler.report_pipes_framework_exception(
                f"{self.__class__.__name__} reader thread",
                sys.exc_info(),
            )
            raise

    def no_messages_debug_text(self) -> str:
        return f"Attempted to read messages from file {self._path}."


class PipesTempFileMessageReader(PipesMessageReader):
    """Message reader that reads messages by tailing an automatically-generated temporary file."""

    def __init__(self, include_stdio_in_messages: bool = False):
        self._include_stdio_in_messages = check.bool_param(
            include_stdio_in_messages, "include_stdio_in_messages"
        )

    @contextmanager
    def read_messages(
        self,
        handler: "PipesMessageHandler",
    ) -> Iterator[PipesParams]:
        """Set up a thread to read streaming messages from the external process by an
        automatically-generated temporary file.

        Args:
            handler (PipesMessageHandler): object to process incoming messages

        Yields:
            PipesParams: A dict of parameters that specifies where a pipes process should write
            pipes protocol messages.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            with PipesFileMessageReader(
                os.path.join(tempdir, _MESSAGE_READER_FILENAME),
                include_stdio_in_messages=self._include_stdio_in_messages,
            ).read_messages(handler) as params:
                yield params

    def no_messages_debug_text(self) -> str:
        return "Attempted to read messages from a local temporary file."


# Time in seconds to wait between attempts when polling for some condition. Default value that is
# used in several places.
DEFAULT_SLEEP_INTERVAL = 1

# Wait up to this many seconds for threads to finish executing during cleanup. Note that this must
# be longer than WAIT_FOR_LOGS_TIMEOUT.
THREAD_WAIT_TIMEOUT = 120

# Buffer period after the external process has completed before we attempt to close out the log
# thread. The purpose of this is to allow an external system to update logs to their final state,
# which may take some time after the external process has completed. This is subtly different than
# `WAIT_FOR_LOGS_TIMEOUT`, which is the amount of time we will wait for targeted log files to exist
# at all. In contrast, `WAIT_FOR_LOGS_AFTER_EXECUTION_SECONDS` is used to guard against the case
# where log files exist but incomplete immediately after the external process has completed.
WAIT_FOR_LOGS_AFTER_EXECUTION_INTERVAL = 10

# Number of seconds to wait after an external process has completed to close
# out the log thread. This is necessary because the log thread contains two
# points at which it waits indefinitely: (1) for the `opened` message to be
# received; (2) for the log files to exist (which may not occur until some time
# after the external process has completed).
WAIT_FOR_LOGS_TIMEOUT = 60


class PipesThreadedMessageReader(PipesMessageReader):
    """A base class for message readers that read messages and logs in background threads.

    Args:
        interval (float): The interval in seconds at which to poll for messages.
        log_readers (Optional[Sequence[PipesLogReader]]): A set of log readers to use to read logs.
    """

    interval: float
    log_readers: dict[str, "PipesLogReader"]
    opened_payload: Optional[PipesOpenedData]
    launched_payload: Optional[PipesLaunchedData]

    def __init__(
        self,
        interval: float = 10,
        log_readers: Optional[Sequence["PipesLogReader"]] = None,
    ):
        self.interval = interval

        self.log_readers = {
            str(i): reader
            for i, reader in enumerate(
                check.opt_sequence_param(log_readers, "log_readers", of_type=PipesLogReader)
            )
        }
        self.opened_payload = None
        self.launched_payload = None

    @contextmanager
    def read_messages(
        self,
        handler: "PipesMessageHandler",
    ) -> Iterator[PipesParams]:
        """Set up a thread to read streaming messages by periodically reading message chunks from a
        target location.

        Args:
            handler (PipesMessageHandler): object to process incoming messages

        Yields:
            PipesParams: A dict of parameters that specifies where a pipes process should write
            pipes protocol message chunks.
        """
        with self.get_params() as params:
            is_session_closed = Event()
            messages_thread = None
            logs_thread = None
            try:
                messages_thread = Thread(
                    target=self._messages_thread,
                    args=(handler, params, is_session_closed),
                    daemon=True,
                )
                messages_thread.start()
                logs_thread = Thread(
                    target=self._logs_thread,
                    args=(handler, params, is_session_closed, messages_thread),
                    daemon=True,
                )
                logs_thread.start()
                yield params
            finally:
                is_session_closed.set()
                if messages_thread:
                    _join_thread(messages_thread, "messages")
                if logs_thread:
                    _join_thread(logs_thread, "logs")

    def on_opened(self, opened_payload: PipesOpenedData) -> None:
        self.opened_payload = opened_payload

    def on_launched(self, launched_payload: PipesLaunchedData) -> None:
        self.launched_payload = launched_payload

    def add_log_reader(self, log_reader: "PipesLogReader") -> None:
        """Can be used to attach extra log readers to the message reader.
        Typically called when the target for reading logs is not known until after the external
        process has started (for example, when the target depends on an external job_id).
        The LogReader will be eventually started by the PipesThreadedMessageReader.
        """
        key = str(len(self.log_readers))
        self.log_readers[key] = log_reader

    @abstractmethod
    def messages_are_readable(self, params: PipesParams) -> bool: ...

    @abstractmethod
    @contextmanager
    def get_params(self) -> Iterator[PipesParams]:
        """Yield a set of parameters to be passed to a message writer in a pipes process.

        Yields:
            PipesParams: A dict of parameters that specifies where a pipes process should write
            pipes protocol message chunks.
        """

    @abstractmethod
    def download_messages(
        self, cursor: Optional[TCursor], params: PipesParams
    ) -> Optional[tuple[TCursor, str]]:
        """Download a chunk of messages from the target location.

        Args:
            cursor (Optional[Any]): Cursor specifying start location from which to download
                messages in a stream. The format of the value varies with the message reader
                implementation. It might be an integer index for a line in a log file, or a
                timestamp for a message in a time-indexed stream.
            params (PipesParams): A dict of parameters that specifies where to download messages from.
        """
        ...

    def _messages_thread(
        self,
        handler: "PipesMessageHandler",
        params: PipesParams,
        is_session_closed: Event,
    ) -> None:
        try:
            start_or_last_download = datetime.datetime.now()
            session_closed_at = None
            cursor = None
            can_read_messages = False

            # main loop to read messages
            # at every step, we:
            # - exit early if we have received the closed message
            # - consume params from the launched_payload if possible
            # - check if we can start reading messages (e.g. log files are available)
            # - download a chunk of messages and process them
            # - if is_session_closed is set, we exit the loop after waiting for WAIT_FOR_LOGS_AFTER_EXECUTION_INTERVAL
            while True:
                # if we have the closed message, we can exit
                # since the message reader has been started and the external process has completed
                if handler.received_closed_message:
                    return

                if not can_read_messages:  # this branch will be executed until we can read messages
                    # check for new params in case they have been updated
                    params = {**params, **(self.launched_payload or {})}
                    can_read_messages = self.messages_are_readable(params)

                now = datetime.datetime.now()
                if (
                    now - start_or_last_download
                ).seconds > self.interval or is_session_closed.is_set():
                    if can_read_messages:
                        start_or_last_download = now
                        result = self.download_messages(cursor, params)
                        if result is not None:
                            cursor, chunk = result
                            for line in chunk.split("\n"):
                                try:
                                    message = json.loads(line)
                                    if PIPES_PROTOCOL_VERSION_FIELD in message.keys():
                                        handler.handle_message(message)
                                except json.JSONDecodeError:
                                    pass

                time.sleep(DEFAULT_SLEEP_INTERVAL)

                if is_session_closed.is_set():
                    if session_closed_at is None:
                        session_closed_at = datetime.datetime.now()

                    # After the external process has completed, we don't want to immediately exit
                    if (
                        datetime.datetime.now() - session_closed_at
                    ).seconds > WAIT_FOR_LOGS_AFTER_EXECUTION_INTERVAL:
                        if not can_read_messages:
                            self._log_unstartable_warning(handler, params)
                        return

        except:
            handler.report_pipes_framework_exception(
                f"{self.__class__.__name__} messages thread",
                sys.exc_info(),
            )
            raise

    def _log_unstartable_warning(self, handler: PipesMessageHandler, params: PipesParams) -> None:
        if self.launched_payload is not None:
            handler._context.log.warning(  # noqa: SLF001
                f"[pipes] Target of {self.__class__.__name__} is not readable after receiving extra params from the external process (`on_launched` has been called)"
            )
        else:
            handler._context.log.warning(  # noqa: SLF001
                f"[pipes] Target of {self.__class__.__name__} is not readable."
            )

    def _logs_thread(
        self,
        handler: "PipesMessageHandler",
        params: PipesParams,
        is_session_closed: Event,
        messages_thread: Thread,
    ) -> None:
        wait_for_logs_start = None

        # Loop over all log readers and start them if the target is readable, which typically means
        # a file exists at the target location. Different execution environments may write logs at
        # different times (e.g., some may write logs periodically during execution, while others may
        # only write logs after the process has completed).
        try:
            unstarted_log_readers = {**self.log_readers}

            while True:
                if self.opened_payload is not None:
                    params = {**params, **self.opened_payload}

                # periodically check for new readers which may be added after the
                # external process has started and add them to the unstarted log readers
                for key in self.log_readers:
                    if key not in unstarted_log_readers:
                        unstarted_log_readers[key] = self.log_readers[key]

                for key in list(unstarted_log_readers.keys()).copy():
                    if unstarted_log_readers[key].target_is_readable(params):
                        reader = unstarted_log_readers.pop(key)
                        reader.start(params, is_session_closed)

                # In some cases logs might not be written out until after the external process has
                # exited. That will leave us in this state, where some log readers have not been
                # started even though the external process is finished. We start a timer and wait
                # for up to WAIT_FOR_LOGS_TIMEOUT seconds for the logs to be written. If they are
                # not written after this amount of time has elapsed, we warn the user and bail.
                if is_session_closed.is_set():
                    if wait_for_logs_start is None:
                        wait_for_logs_start = datetime.datetime.now()

                    if not unstarted_log_readers:
                        return
                    elif (
                        unstarted_log_readers
                        and (datetime.datetime.now() - wait_for_logs_start).seconds
                        > WAIT_FOR_LOGS_TIMEOUT
                    ):
                        for key, log_reader in unstarted_log_readers.items():
                            warnings.warn(
                                log_reader.with_debug_info(
                                    f"[pipes] Attempted to read log for reader {log_reader.name} but log was"
                                    f" still not written {WAIT_FOR_LOGS_TIMEOUT} seconds after session close. Abandoning reader {key}."
                                )
                            )

                        return

                time.sleep(DEFAULT_SLEEP_INTERVAL)
        except Exception:
            handler.report_pipes_framework_exception(
                f"{self.__class__.__name__} logs thread",
                sys.exc_info(),
            )
            raise
        finally:
            for log_reader in self.log_readers.values():
                if log_reader.is_running():
                    log_reader.stop()


class PipesBlobStoreMessageReader(PipesThreadedMessageReader):
    """Message reader that reads a sequence of message chunks written by an external process into a
    blob store such as S3, Azure blob storage, or GCS.

    The reader maintains a counter, starting at 1, that is synchronized with a message writer in
    some pipes process. The reader starts a thread that periodically attempts to read a chunk
    indexed by the counter at some location expected to be written by the pipes process. The chunk
    should be a file with each line corresponding to a JSON-encoded pipes message. When a chunk is
    successfully read, the messages are processed and the counter is incremented. The
    :py:class:`PipesBlobStoreMessageWriter` on the other end is expected to similarly increment a
    counter (starting from 1) on successful write, keeping counters on the read and write end in
    sync.

    If `log_readers` is passed, the message reader will start the passed log readers when the
    `opened` message is received from the external process.

    Args:
        interval (float): interval in seconds between attempts to download a chunk
        log_readers (Optional[Sequence[PipesLogReader]]): A set of log readers to use to read logs.
    """

    counter: int

    def __init__(
        self,
        interval: float = 10,
        log_readers: Optional[Sequence["PipesLogReader"]] = None,
    ):
        super().__init__(interval=interval, log_readers=log_readers)

        self.counter = 1

    @abstractmethod
    def download_messages_chunk(self, index: int, params: PipesParams) -> Optional[str]:
        ...
        # historical reasons, keeping the original interface of PipesBlobStoreMessageReader

    def download_messages(
        self, cursor: Optional[int], params: PipesParams
    ) -> Optional[tuple[int, str]]:
        # mapping new interface to the old one
        # the old interface isn't using the cursor parameter, instead, it keeps track of counter in the "counter" attribute
        chunk = self.download_messages_chunk(self.counter, params)
        if chunk:
            self.counter += 1
            return self.counter, chunk


class PipesLogReader(ABC):
    @abstractmethod
    def start(self, params: PipesParams, is_session_closed: Event) -> None: ...

    @abstractmethod
    def stop(self) -> None: ...

    @abstractmethod
    def is_running(self) -> bool: ...

    @abstractmethod
    def target_is_readable(self, params: PipesParams) -> bool: ...

    @property
    def name(self) -> str:
        """Override this to distinguish different log readers in error messages."""
        return self.__class__.__name__

    @property
    def debug_info(self) -> Optional[str]:
        """An optional message containing debug information about the log reader.

        It will be included in error messages when the log reader fails to start or throws an exception.
        """

    def with_debug_info(self, message: str) -> str:
        """Helper method to include debug info in error messages."""
        return f"{message}\nDebug info: {self.debug_info}" if self.debug_info else message


class PipesChunkedLogReader(PipesLogReader):
    """Reader for reading stdout/stderr logs from a blob store such as S3, Azure blob storage, or GCS.

    Args:
        interval (float): interval in seconds between attempts to download a chunk.
        target_stream (IO[str]): The stream to which to write the logs. Typcially `sys.stdout` or `sys.stderr`.
        debug_info (Optional[str]): An optional message containing debug information about the log reader.
    """

    def __init__(
        self, *, interval: float = 10, target_stream: IO[str], debug_info: Optional[str] = None
    ):
        self.interval = interval
        self.target_stream = target_stream
        self.thread: Optional[Thread] = None
        self._debug_info = debug_info

    @property
    def debug_info(self) -> Optional[str]:
        return self._debug_info

    @abstractmethod
    def download_log_chunk(self, params: PipesParams) -> Optional[str]: ...

    def start(self, params: PipesParams, is_session_closed: Event) -> None:
        self.thread = Thread(target=self._reader_thread, args=(params, is_session_closed))
        self.thread.start()

    def stop(self) -> None:
        if self.thread is None:
            raise DagsterInvariantViolationError(
                "Attempted to wait for log reader to finish, but it was never started."
            )
        _join_thread(self.thread, self.name)

    def is_running(self) -> bool:
        return self.thread is not None and self.thread.is_alive()

    def _reader_thread(
        self,
        params: PipesParams,
        is_session_closed: Event,
    ) -> None:
        start_or_last_download = datetime.datetime.now()
        after_execution_time_start = None
        while True:
            now = datetime.datetime.now()
            if (now - start_or_last_download).seconds > self.interval or is_session_closed.is_set():
                start_or_last_download = now
                chunk = self.download_log_chunk(params)
                if chunk:
                    self.target_stream.write(chunk)

                # After execution is complete, we don't want to immediately exit, because it is
                # possible the external system will take some time to flush logs to the external
                # storage system. Only exit after WAIT_FOR_LOGS_AFTER_EXECUTION_INTERVAL seconds
                # have elapsed.
                elif is_session_closed.is_set():
                    if after_execution_time_start is None:
                        after_execution_time_start = datetime.datetime.now()
                    elif (
                        datetime.datetime.now() - after_execution_time_start
                    ).seconds > WAIT_FOR_LOGS_AFTER_EXECUTION_INTERVAL:
                        break
                time.sleep(self.interval)


def _join_thread(thread: Thread, thread_name: str) -> None:
    thread.join(timeout=THREAD_WAIT_TIMEOUT)
    if thread.is_alive():
        raise DagsterPipesExecutionError(f"Timed out waiting for {thread_name} thread to finish.")


def forward_only_logs_to_file(log_line: str, file: IO[str]):
    """Will write the log line to the file if it is not a Pipes message."""
    try:
        message = json.loads(log_line)
        if PIPES_PROTOCOL_VERSION_FIELD in message.keys():
            return
        else:
            file.writelines((log_line, "\n"))
    except Exception:
        file.writelines((log_line, "\n"))


def extract_message_or_forward_to_file(
    handler: "PipesMessageHandler", log_line: str, file: IO[str]
):
    # exceptions as control flow, you love to see it
    try:
        message = json.loads(log_line)
        if PIPES_PROTOCOL_VERSION_FIELD in message.keys():
            handler.handle_message(message)
        else:
            file.writelines((log_line, "\n"))
    except Exception:
        # move non-message logs in to file for compute log capture
        file.writelines((log_line, "\n"))


def extract_message_or_forward_to_stdout(handler: "PipesMessageHandler", log_line: str):
    extract_message_or_forward_to_file(handler=handler, log_line=log_line, file=sys.stdout)


_FAIL_TO_YIELD_ERROR_MESSAGE = (
    "Did you forget to `yield from pipes_session.get_results()` or `return"
    " <PipesClient>.run(...).get_results`? If using `open_pipes_session`,"
    " `pipes_session.get_results` should be called once after the `open_pipes_session` block has"
    " exited to yield any remaining buffered results via `<PipesSession>.get_results()`."
    " If using `<PipesClient>.run`, you should always return"
    " `<PipesClient>.run(...).get_results()` or `<PipesClient>.run(...).get_materialize_result()`."
)


@contextmanager
def open_pipes_session(
    context: Union[OpExecutionContext, AssetExecutionContext],
    context_injector: PipesContextInjector,
    message_reader: PipesMessageReader,
    extras: Optional[PipesExtras] = None,
) -> Iterator[PipesSession]:
    """Context manager that opens and closes a pipes session.

    This context manager should be used to wrap the launch of an external process using the pipe
    protocol to report results back to Dagster. The yielded :py:class:`PipesSession` should be used
    to (a) obtain the environment variables that need to be provided to the external process; (b)
    access results streamed back from the external process.

    This method is an alternative to :py:class:`PipesClient` subclasses for users who want more
    control over how pipes processes are launched. When using `open_pipes_session`, it is the user's
    responsibility to inject the message reader and context injector parameters available on the
    yielded `PipesSession` and pass them to the appropriate API when launching the external process.
    Typically these parameters should be set as environment variables.


    Args:
        context (Union[OpExecutionContext, AssetExecutionContext]): The context for the current op/asset execution.
        context_injector (PipesContextInjector): The context injector to use to inject context into the external process.
        message_reader (PipesMessageReader): The message reader to use to read messages from the external process.
        extras (Optional[PipesExtras]): Optional extras to pass to the external process via the injected context.

    Yields:
        PipesSession: Interface for interacting with the external process.

    .. code-block:: python

        import subprocess
        from dagster import open_pipes_session

        extras = {"foo": "bar"}

        @asset
        def ext_asset(context: AssetExecutionContext):
            with open_pipes_session(
                context=context,
                extras={"foo": "bar"},
                context_injector=PipesTempFileContextInjector(),
                message_reader=PipesTempFileMessageReader(),
            ) as pipes_session:
                subprocess.Popen(
                    ["/bin/python", "/path/to/script.py"],
                    env={**pipes_session.get_bootstrap_env_vars()}
                )
                while process.poll() is None:
                    yield from pipes_session.get_results()

            yield from pipes_session.get_results()
    """
    # if we are in the context of an asset, set up a defensive check to ensure the event stream got returned
    if context.has_assets_def:
        context.set_requires_typed_event_stream(error_message=_FAIL_TO_YIELD_ERROR_MESSAGE)

    context_data = build_external_execution_context_data(context, extras)
    message_handler = PipesMessageHandler(context, message_reader)
    try:
        with (
            context_injector.inject_context(context_data) as ci_params,
            message_handler.handle_messages() as mr_params,
        ):
            yield PipesSession(
                context_data=context_data,
                message_handler=message_handler,
                context_injector_params=ci_params,
                message_reader_params=mr_params,
                context=context,
            )
    finally:
        if not message_handler.received_opened_message:
            context.log.warning(
                "[pipes] did not receive any messages from external process. Check stdout / stderr"
                " logs from the external process if"
                f" possible.\n{context_injector.__class__.__name__}:"
                f" {context_injector.no_messages_debug_text()}\n{message_reader.__class__.__name__}:"
                f" {message_reader.no_messages_debug_text()}\n"
            )
        elif not message_handler.received_closed_message:
            context.log.warning(
                "[pipes] did not receive closed message from external process. Buffered messages"
                " may have been discarded without being delivered. Use `open_dagster_pipes` as a"
                " context manager (a with block) to ensure that cleanup is successfully completed."
                " If that is not possible, manually call `PipesContext.close()` before process"
                " exit."
            )



import os
from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Optional
from urllib.parse import urljoin, urlparse

import sqlalchemy as db
from sqlalchemy.engine import Connection
from sqlalchemy.pool import NullPool
from typing_extensions import Self

from dagster import (
    StringSource,
    _check as check,
)
from dagster._config.config_schema import UserConfigSchema
from dagster._core.storage.runs.schema import (
    InstanceInfo,
    RunsTable,
    RunStorageSqlMetadata,
    RunTagsTable,
)
from dagster._core.storage.runs.sql_run_storage import SqlRunStorage
from dagster._core.storage.sql import (
    AlembicVersion,
    check_alembic_revision,
    create_engine,
    get_alembic_config,
    run_alembic_downgrade,
    run_alembic_upgrade,
    safe_commit,
    stamp_alembic_rev,
)
from dagster._core.storage.sqlite import (
    LAST_KNOWN_STAMPED_SQLITE_ALEMBIC_REVISION,
    create_db_conn_string,
)
from dagster._serdes import ConfigurableClass, ConfigurableClassData
from dagster._utils import mkdir_p

if TYPE_CHECKING:
    from dagster._core.storage.sqlite_storage import SqliteStorageConfig
MINIMUM_SQLITE_BUCKET_VERSION = [3, 25, 0]


class SqliteRunStorage(SqlRunStorage, ConfigurableClass):
    """SQLite-backed run storage.

    Users should not directly instantiate this class; it is instantiated by internal machinery when
    ``dagster-webserver`` and ``dagster-graphql`` load, based on the values in the ``dagster.yaml`` file in
    ``$DAGSTER_HOME``. Configuration of this class should be done by setting values in that file.

    This is the default run storage when none is specified in the ``dagster.yaml``.

    To explicitly specify SQLite for run storage, you can add a block such as the following to your
    ``dagster.yaml``:

    .. code-block:: YAML

        run_storage:
          module: dagster._core.storage.runs
          class: SqliteRunStorage
          config:
            base_dir: /path/to/dir

    The ``base_dir`` param tells the run storage where on disk to store the database.
    """

    def __init__(self, conn_string: str, inst_data: Optional[ConfigurableClassData] = None):
        check.str_param(conn_string, "conn_string")
        self._conn_string = conn_string
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {"base_dir": StringSource}

    @classmethod
    def from_config_value(
        cls, inst_data: Optional[ConfigurableClassData], config_value: "SqliteStorageConfig"
    ) -> "SqliteRunStorage":
        return SqliteRunStorage.from_local(inst_data=inst_data, **config_value)

    @classmethod
    def from_local(cls, base_dir: str, inst_data: Optional[ConfigurableClassData] = None) -> Self:
        check.str_param(base_dir, "base_dir")
        mkdir_p(base_dir)
        conn_string = create_db_conn_string(base_dir, "runs")
        engine = create_engine(conn_string, poolclass=NullPool)
        alembic_config = get_alembic_config(__file__)

        should_mark_indexes = False
        with engine.connect() as connection:
            db_revision, head_revision = check_alembic_revision(alembic_config, connection)
            table_names = db.inspect(engine).get_table_names()
            if not (db_revision and head_revision):
                if "runs" in table_names:
                    # The runs table exists but the alembic version table does not. This means that the SQLite db was
                    # initialized with SQLAlchemy 2.0 before https://github.com/dagster-io/dagster/pull/25740 was merged.
                    # We should pin the alembic revision to the last known stamped revision before we unpinned SQLAlchemy 2.0
                    # This should be safe because we have guarded all known migrations since then.
                    rev_to_stamp = LAST_KNOWN_STAMPED_SQLITE_ALEMBIC_REVISION
                else:
                    should_mark_indexes = True
                    rev_to_stamp = "head"

                RunStorageSqlMetadata.create_all(engine)
                connection.execute(db.text("PRAGMA journal_mode=WAL;"))
                stamp_alembic_rev(alembic_config, connection, rev=rev_to_stamp)
                safe_commit(connection)

            table_names = db.inspect(engine).get_table_names()
            if "instance_info" not in table_names:
                InstanceInfo.create(engine)

        run_storage = cls(conn_string, inst_data)

        if should_mark_indexes:
            run_storage.migrate()
            run_storage.optimize()

        return run_storage

    @contextmanager
    def connect(self) -> Iterator[Connection]:
        engine = create_engine(self._conn_string, poolclass=NullPool)
        with engine.connect() as conn:
            with conn.begin():
                yield conn

    def _alembic_upgrade(self, rev: str = "head") -> None:
        alembic_config = get_alembic_config(__file__)
        with self.connect() as conn:
            run_alembic_upgrade(alembic_config, conn, rev=rev)

    def _alembic_downgrade(self, rev: str = "head") -> None:
        alembic_config = get_alembic_config(__file__)
        with self.connect() as conn:
            run_alembic_downgrade(alembic_config, conn, rev=rev)

    def upgrade(self) -> None:
        self._check_for_version_066_migration_and_perform()
        self._alembic_upgrade()

    # In version 0.6.6, we changed the layout of the of the sqllite dbs on disk
    # to move from the root of DAGSTER_HOME/runs.db to DAGSTER_HOME/history/runs.bd
    # This function checks for that condition and does the move
    def _check_for_version_066_migration_and_perform(self) -> None:
        old_conn_string = "sqlite://" + urljoin(urlparse(self._conn_string).path, "../runs.db")
        path_to_old_db = urlparse(old_conn_string).path
        # sqlite URLs look like `sqlite:///foo/bar/baz on Unix/Mac` but on Windows they look like
        # `sqlite:///D:/foo/bar/baz` (or `sqlite:///D:\foo\bar\baz`)
        if os.name == "nt":
            path_to_old_db = path_to_old_db.lstrip("/")
        if os.path.exists(path_to_old_db):
            old_storage = SqliteRunStorage(old_conn_string)
            old_runs = old_storage.get_runs()
            for run in old_runs:
                self.add_run(run)
            os.unlink(path_to_old_db)

    def delete_run(self, run_id: str) -> None:
        """Override the default sql delete run implementation until we can get full
        support on cascading deletes.
        """
        check.str_param(run_id, "run_id")
        remove_tags = db.delete(RunTagsTable).where(RunTagsTable.c.run_id == run_id)
        remove_run = db.delete(RunsTable).where(RunsTable.c.run_id == run_id)
        with self.connect() as conn:
            conn.execute(remove_tags)
            conn.execute(remove_run)

    def alembic_version(self) -> AlembicVersion:
        alembic_config = get_alembic_config(__file__)
        with self.connect() as conn:
            return check_alembic_revision(alembic_config, conn)



import json
from collections.abc import Mapping
from typing import Any, Optional

from dagster._config.snap import ConfigSchemaSnapshot, ConfigTypeSnap
from dagster._utils.yaml_utils import dump_run_config_yaml


def _safe_json_loads(json_str: Optional[str]) -> object:
    try:
        return json.loads(json_str) if json_str else None
    except json.JSONDecodeError:
        return None


PRIORITY_CONFIG_KEYS = ("ops", "resources")


def _filter_empty_dicts(to_filter: Any) -> Any:
    if not isinstance(to_filter, Mapping):
        return to_filter
    else:
        filtered_dict = {k: _filter_empty_dicts(v) for k, v in to_filter.items()}
        return {k: v for k, v in filtered_dict.items() if v is not None and v != {}}


def _cleanup_run_config_dict(run_config_dict: Any) -> Any:
    """Performs cleanup of the run config dict to remove empty dicts and strip the default executor
    config if it has not been overridden, to make the output more readable.
    """
    return _filter_empty_dicts(run_config_dict)


def default_values_yaml_from_type_snap(
    snapshot: ConfigSchemaSnapshot,
    type_snap: ConfigTypeSnap,
) -> str:
    """Returns a YAML representation of the default values for the given type snap."""
    run_config_dict = _cleanup_run_config_dict(default_values_from_type_snap(type_snap, snapshot))

    # Sort the keys so that the output begins with the most useful keys (ops, resources)
    # We use a dict rather than an OrderedDict because in Py3.7+ the order of keys in a dict
    # is guaranteed to be insertion order and because yaml.dump() does not natively support
    # OrderedDicts
    run_config_dict_sorted: Mapping[str, Any] = dict(
        (k, run_config_dict.get(k))
        for k in [*PRIORITY_CONFIG_KEYS, *run_config_dict.keys()]
        if k in run_config_dict
    )
    return dump_run_config_yaml(run_config_dict_sorted, sort_keys=False)


def default_values_from_type_snap(type_snap: ConfigTypeSnap, snapshot: ConfigSchemaSnapshot) -> Any:
    """Given a type snap and a snapshot, returns a dictionary of default values for the type
    snap, recursively assembling a default if the type snap does not have a default value
    explicitly set.
    """
    if not type_snap.fields:
        return {}

    defaults_by_field = {}
    for field_name in type_snap.field_names:
        field = type_snap.get_field(field_name)

        default_value_as_json = field.default_value_as_json_str
        field_snap = (
            snapshot.get_config_snap(field.type_key)
            if snapshot.has_config_snap(field.type_key)
            else None
        )

        # First, we try to get the default value from the field itself
        # this is usually only set for primitive field types with user-supplied defaults
        if default_value_as_json:
            defaults_by_field[field_name] = _safe_json_loads(default_value_as_json)
        # If there is no default value on the field, if the field has child fields, we recurse
        # to assemble the default values for the child fields
        elif field_snap and field_snap.fields:
            defaults_by_field[field_name] = default_values_from_type_snap(field_snap, snapshot)

    return defaults_by_field



from collections.abc import Sequence
from typing import TYPE_CHECKING

import dagster._check as check
from dagster._core.errors import DagsterUserCodeProcessError
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation.external_data import (
    PartitionConfigSnap,
    PartitionExecutionErrorSnap,
    PartitionNamesSnap,
    PartitionSetExecutionParamSnap,
    PartitionTagsSnap,
    partition_set_snap_name_for_job_name,
)
from dagster._core.remote_representation.handle import RepositoryHandle
from dagster._grpc.types import PartitionArgs, PartitionNamesArgs, PartitionSetExecutionParamArgs
from dagster._serdes import deserialize_value

if TYPE_CHECKING:
    from dagster._grpc.client import DagsterGrpcClient


def sync_get_external_partition_names_grpc(
    api_client: "DagsterGrpcClient",
    repository_handle: RepositoryHandle,
    job_name: str,
) -> PartitionNamesSnap:
    from dagster._grpc.client import DagsterGrpcClient

    check.inst_param(api_client, "api_client", DagsterGrpcClient)
    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(job_name, "job_name")
    repository_origin = repository_handle.get_remote_origin()
    result = deserialize_value(
        api_client.external_partition_names(
            partition_names_args=PartitionNamesArgs(
                repository_origin=repository_origin,
                job_name=job_name,
                partition_set_name=partition_set_snap_name_for_job_name(job_name),
            ),
        ),
        (PartitionNamesSnap, PartitionExecutionErrorSnap),
    )
    if isinstance(result, PartitionExecutionErrorSnap):
        raise DagsterUserCodeProcessError.from_error_info(result.error)

    return result


def sync_get_external_partition_config_grpc(
    api_client: "DagsterGrpcClient",
    repository_handle: RepositoryHandle,
    job_name: str,
    partition_name: str,
    instance: DagsterInstance,
) -> PartitionConfigSnap:
    from dagster._grpc.client import DagsterGrpcClient

    check.inst_param(api_client, "api_client", DagsterGrpcClient)
    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(job_name, "job_name")
    check.str_param(partition_name, "partition_name")
    repository_origin = repository_handle.get_remote_origin()
    result = deserialize_value(
        api_client.external_partition_config(
            partition_args=PartitionArgs(
                repository_origin=repository_origin,
                job_name=job_name,
                partition_set_name=partition_set_snap_name_for_job_name(job_name),
                partition_name=partition_name,
                instance_ref=instance.get_ref(),
            ),
        ),
        (PartitionConfigSnap, PartitionExecutionErrorSnap),
    )
    if isinstance(result, PartitionExecutionErrorSnap):
        raise DagsterUserCodeProcessError.from_error_info(result.error)

    return result


def sync_get_external_partition_tags_grpc(
    api_client: "DagsterGrpcClient",
    repository_handle: RepositoryHandle,
    job_name: str,
    partition_name: str,
    instance: DagsterInstance,
) -> PartitionTagsSnap:
    from dagster._grpc.client import DagsterGrpcClient

    check.inst_param(api_client, "api_client", DagsterGrpcClient)
    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(job_name, "job_name")
    check.str_param(partition_name, "partition_name")

    repository_origin = repository_handle.get_remote_origin()
    result = deserialize_value(
        api_client.external_partition_tags(
            partition_args=PartitionArgs(
                repository_origin=repository_origin,
                job_name=job_name,
                partition_set_name=partition_set_snap_name_for_job_name(job_name),
                partition_name=partition_name,
                instance_ref=instance.get_ref(),
            ),
        ),
        (PartitionTagsSnap, PartitionExecutionErrorSnap),
    )
    if isinstance(result, PartitionExecutionErrorSnap):
        raise DagsterUserCodeProcessError.from_error_info(result.error)

    return result


def sync_get_external_partition_set_execution_param_data_grpc(
    api_client: "DagsterGrpcClient",
    repository_handle: RepositoryHandle,
    partition_set_name: str,
    partition_names: Sequence[str],
    instance: DagsterInstance,
) -> PartitionSetExecutionParamSnap:
    from dagster._grpc.client import DagsterGrpcClient

    check.inst_param(api_client, "api_client", DagsterGrpcClient)
    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(partition_set_name, "partition_set_name")
    check.sequence_param(partition_names, "partition_names", of_type=str)

    repository_origin = repository_handle.get_remote_origin()

    result = deserialize_value(
        api_client.external_partition_set_execution_params(
            partition_set_execution_param_args=PartitionSetExecutionParamArgs(
                repository_origin=repository_origin,
                partition_set_name=partition_set_name,
                partition_names=partition_names,
                instance_ref=instance.get_ref(),
            ),
        ),
        (PartitionSetExecutionParamSnap, PartitionExecutionErrorSnap),
    )
    if isinstance(result, PartitionExecutionErrorSnap):
        raise DagsterUserCodeProcessError.from_error_info(result.error)

    return result



import keyword
import os
import re
from collections.abc import Iterable, Mapping, Sequence
from glob import glob
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union, cast

import yaml

import dagster._check as check
from dagster._core.definitions.asset_key import AssetCheckKey, EntityKey
from dagster._core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster._core.utils import is_valid_email
from dagster._utils.warnings import deprecation_warning, disable_dagster_warnings
from dagster._utils.yaml_utils import merge_yaml_strings, merge_yamls

DEFAULT_OUTPUT = "result"
DEFAULT_GROUP_NAME = "default"  # asset group_name used when none is provided
DEFAULT_IO_MANAGER_KEY = "io_manager"

DISALLOWED_NAMES = set(
    [
        "context",
        "conf",
        "config",
        "meta",
        "arg_dict",
        "dict",
        "input_arg_dict",
        "output_arg_dict",
        "int",
        "str",
        "float",
        "bool",
        "input",
        "output",
        "type",
    ]
    + list(keyword.kwlist)  # just disallow all python keywords
)

INVALID_NAME_CHARS = r"[^A-Za-z0-9_]"
VALID_NAME_REGEX_STR = r"^[A-Za-z0-9_]+$"
VALID_NAME_REGEX = re.compile(VALID_NAME_REGEX_STR)

INVALID_TITLE_CHARACTERS_REGEX_STR = r"[\%\*\"]"
INVALID_TITLE_CHARACTERS_REGEX = re.compile(INVALID_TITLE_CHARACTERS_REGEX_STR)
MAX_TITLE_LENGTH = 100

if TYPE_CHECKING:
    from dagster._core.definitions.asset_key import AssetKey
    from dagster._core.definitions.asset_selection import AssetSelection
    from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
    from dagster._core.definitions.base_asset_graph import BaseAssetGraph
    from dagster._core.definitions.declarative_automation.automation_condition import (
        AutomationCondition,
    )
    from dagster._core.definitions.sensor_definition import SensorDefinition
    from dagster._core.remote_representation.external import RemoteSensor


class NoValueSentinel:
    """Sentinel value to distinguish unset from None."""


def has_valid_name_chars(name: str) -> bool:
    return bool(VALID_NAME_REGEX.match(name))


def check_valid_name(name: str, allow_list: Optional[list[str]] = None) -> str:
    check.str_param(name, "name")

    if allow_list and name in allow_list:
        return name

    if name in DISALLOWED_NAMES:
        raise DagsterInvalidDefinitionError(
            f'"{name}" is not a valid name in Dagster. It conflicts with a Dagster or python'
            " reserved keyword."
        )

    check_valid_chars(name)

    check.invariant(is_valid_name(name))
    return name


def check_valid_chars(name: str):
    if not has_valid_name_chars(name):
        raise DagsterInvalidDefinitionError(
            f'"{name}" is not a valid name in Dagster. Names must be in regex'
            f" {VALID_NAME_REGEX_STR}."
        )


def is_valid_name(name: str) -> bool:
    check.str_param(name, "name")

    return name not in DISALLOWED_NAMES and has_valid_name_chars(name)


def is_valid_title_and_reason(title: Optional[str]) -> tuple[bool, Optional[str]]:
    check.opt_str_param(title, "title")

    if title is None:
        return True, None

    if len(title) > MAX_TITLE_LENGTH:
        return (
            False,
            f'"{title}" ({len(title)} characters) is not a valid title in Dagster. Titles must not be longer than {MAX_TITLE_LENGTH}.',
        )

    if not is_valid_title_chars(title):
        return (
            False,
            f'"{title}" is not a valid title in Dagster. Titles must not contain regex {INVALID_TITLE_CHARACTERS_REGEX_STR}.',
        )

    return True, None


def check_valid_title(title: Optional[str]) -> Optional[str]:
    """A title is distinguished from a name in that the title is a descriptive string meant for display in the UI.
    It is not used as an identifier for an object.
    """
    is_valid, reason = is_valid_title_and_reason(title)
    if not is_valid:
        raise DagsterInvariantViolationError(reason)

    return title


def is_valid_title(title: Optional[str]) -> bool:
    return is_valid_title_and_reason(title)[0]


def is_valid_title_chars(title: str):
    return not bool(INVALID_TITLE_CHARACTERS_REGEX.search(title))


def _kv_str(key: object, value: object) -> str:
    return f'{key}="{value!r}"'


def struct_to_string(name: str, **kwargs: object) -> str:
    # Sort the kwargs to ensure consistent representations across Python versions
    props_str = ", ".join([_kv_str(key, value) for key, value in sorted(kwargs.items())])
    return f"{name}({props_str})"


def validate_asset_owner(owner: str, key: "AssetKey") -> None:
    if not is_valid_asset_owner(owner):
        raise DagsterInvalidDefinitionError(
            f"Invalid owner '{owner}' for asset '{key}'. Owner must be an email address or a team "
            "name prefixed with 'team:'."
        )


def is_valid_asset_owner(owner: str) -> bool:
    return is_valid_email(owner) or (owner.startswith("team:") and len(owner) > 5)


def validate_group_name(group_name: Optional[str]) -> None:
    """Ensures a string name is valid and returns a default if no name provided."""
    if group_name:
        check_valid_chars(group_name)
    elif group_name == "":
        raise DagsterInvalidDefinitionError(
            "Empty asset group name was provided, which is not permitted. "
            "Set group_name=None to use the default group_name or set non-empty string"
        )


def normalize_group_name(group_name: Optional[str]) -> str:
    """Ensures a string name is valid and returns a default if no name provided."""
    validate_group_name(group_name)
    return group_name or DEFAULT_GROUP_NAME


def config_from_files(config_files: Sequence[str]) -> Mapping[str, Any]:
    """Constructs run config from YAML files.

    Args:
        config_files (List[str]): List of paths or glob patterns for yaml files
            to load and parse as the run config.

    Returns:
        Dict[str, Any]: A run config dictionary constructed from provided YAML files.

    Raises:
        FileNotFoundError: When a config file produces no results
        DagsterInvariantViolationError: When one of the YAML files is invalid and has a parse
            error.
    """
    config_files = check.opt_sequence_param(config_files, "config_files")

    filenames = []
    for file_glob in config_files or []:
        globbed_files = glob(file_glob)
        if not globbed_files:
            raise DagsterInvariantViolationError(
                f'File or glob pattern "{file_glob}" for "config_files" produced no results.'
            )

        filenames += [os.path.realpath(globbed_file) for globbed_file in globbed_files]

    try:
        run_config = merge_yamls(filenames)
    except yaml.YAMLError as err:
        raise DagsterInvariantViolationError(
            f"Encountered error attempting to parse yaml. Parsing files {filenames} "
            f"loaded by file/patterns {config_files}."
        ) from err

    return check.is_dict(cast(dict[str, object], run_config), key_type=str)


def config_from_yaml_strings(yaml_strings: Sequence[str]) -> Mapping[str, Any]:
    """Static constructor for run configs from YAML strings.

    Args:
        yaml_strings (List[str]): List of yaml strings to parse as the run config.

    Returns:
        Dict[Str, Any]: A run config dictionary constructed from the provided yaml strings

    Raises:
        DagsterInvariantViolationError: When one of the YAML documents is invalid and has a
            parse error.
    """
    yaml_strings = check.sequence_param(yaml_strings, "yaml_strings", of_type=str)

    try:
        run_config = merge_yaml_strings(yaml_strings)
    except yaml.YAMLError as err:
        raise DagsterInvariantViolationError(
            f"Encountered error attempting to parse yaml. Parsing YAMLs {yaml_strings} "
        ) from err

    return check.is_dict(cast(dict[str, object], run_config), key_type=str)


def config_from_pkg_resources(pkg_resource_defs: Sequence[tuple[str, str]]) -> Mapping[str, Any]:
    """Load a run config from a package resource, using :py:func:`pkg_resources.resource_string`.

    Example:
        .. code-block:: python

            config_from_pkg_resources(
                pkg_resource_defs=[
                    ('dagster_examples.airline_demo.environments', 'local_base.yaml'),
                    ('dagster_examples.airline_demo.environments', 'local_warehouse.yaml'),
                ],
            )


    Args:
        pkg_resource_defs (List[(str, str)]): List of pkg_resource modules/files to
            load as the run config.

    Returns:
        Dict[Str, Any]: A run config dictionary constructed from the provided yaml strings

    Raises:
        DagsterInvariantViolationError: When one of the YAML documents is invalid and has a
            parse error.
    """
    import pkg_resources  # expensive, import only on use

    pkg_resource_defs = check.sequence_param(pkg_resource_defs, "pkg_resource_defs", of_type=tuple)

    try:
        yaml_strings = [
            pkg_resources.resource_string(*pkg_resource_def).decode("utf-8")
            for pkg_resource_def in pkg_resource_defs
        ]
    except (ModuleNotFoundError, FileNotFoundError, UnicodeDecodeError) as err:
        raise DagsterInvariantViolationError(
            "Encountered error attempting to parse yaml. Loading YAMLs from "
            f"package resources {pkg_resource_defs}."
        ) from err

    return config_from_yaml_strings(yaml_strings=yaml_strings)


def resolve_automation_condition(
    automation_condition: Optional["AutomationCondition"],
    auto_materialize_policy: Optional["AutoMaterializePolicy"],
) -> Optional["AutomationCondition"]:
    if auto_materialize_policy is not None:
        deprecation_warning(
            "Parameter `auto_materialize_policy`",
            "1.9",
            additional_warn_text="Use `automation_condition` instead.",
        )
        if automation_condition is not None:
            raise DagsterInvariantViolationError(
                "Cannot supply both `automation_condition` and `auto_materialize_policy`"
            )
        return auto_materialize_policy.to_automation_condition()
    else:
        return automation_condition


T = TypeVar("T")


def dedupe_object_refs(objects: Optional[Iterable[T]]) -> Sequence[T]:
    """Dedupe definitions by reference equality."""
    return list({id(obj): obj for obj in objects}.values()) if objects is not None else []


def get_default_automation_condition_sensor(
    sensors: Sequence["SensorDefinition"],
    asset_graph: "BaseAssetGraph",
) -> Optional["SensorDefinition"]:
    """Given a list of existing sensors, adds an AutomationConditionSensorDefinition with name
    `default_automation_condition_sensor` that targets all assets/asset_checks that have an
    automation_condition and are not targeted by an existing AutomationConditionSensorDefinition
    if any such untargeted assets/asset_checks exist.
    """
    from dagster._core.definitions.automation_condition_sensor_definition import (
        DEFAULT_AUTOMATION_CONDITION_SENSOR_NAME,
        AutomationConditionSensorDefinition,
    )

    with disable_dagster_warnings():
        sensor_selection = get_default_automation_condition_sensor_selection(sensors, asset_graph)
        if sensor_selection:
            return AutomationConditionSensorDefinition(
                DEFAULT_AUTOMATION_CONDITION_SENSOR_NAME, target=sensor_selection
            )

    return None


def get_default_automation_condition_sensor_selection(
    sensors: Sequence[Union["SensorDefinition", "RemoteSensor"]], asset_graph: "BaseAssetGraph"
) -> Optional["AssetSelection"]:
    from dagster._core.definitions.asset_selection import AssetSelection
    from dagster._core.definitions.sensor_definition import SensorType

    automation_condition_sensors = sorted(
        (
            s
            for s in sensors
            if s.sensor_type in (SensorType.AUTO_MATERIALIZE, SensorType.AUTOMATION)
        ),
        key=lambda s: s.name,
    )

    automation_condition_keys = set()
    for k in asset_graph.materializable_asset_keys | asset_graph.asset_check_keys:
        if asset_graph.get(k).automation_condition is not None:
            automation_condition_keys.add(k)

    has_auto_observe_keys = False
    for k in asset_graph.observable_asset_keys:
        if (
            # for backcompat, treat auto-observe assets as if they have a condition
            asset_graph.get(k).automation_condition is not None
            or asset_graph.get(k).auto_observe_interval_minutes is not None
        ):
            has_auto_observe_keys = True
            automation_condition_keys.add(k)

    # get the set of keys that are handled by an existing sensor
    covered_keys: set[EntityKey] = set()
    for sensor in automation_condition_sensors:
        selection = check.not_none(sensor.asset_selection)
        covered_keys = covered_keys.union(
            selection.resolve(asset_graph) | selection.resolve_checks(asset_graph)
        )

    default_sensor_keys = automation_condition_keys - covered_keys
    if len(default_sensor_keys) > 0:
        # Use AssetSelection.all if the default sensor is the only sensor - otherwise
        # enumerate the assets that are not already included in some other
        # non-default sensor
        default_sensor_asset_selection = AssetSelection.all(include_sources=has_auto_observe_keys)

        # if there are any asset checks, include checks in the selection
        if any(isinstance(k, AssetCheckKey) for k in default_sensor_keys):
            default_sensor_asset_selection |= AssetSelection.all_asset_checks()

        # remove any selections that are already covered
        for sensor in automation_condition_sensors:
            default_sensor_asset_selection = default_sensor_asset_selection - check.not_none(
                sensor.asset_selection
            )
        return default_sensor_asset_selection
    # no additional sensor required
    else:
        return None



import logging
import os
from enum import Enum
from typing import Optional, TypedDict

from dagster._time import get_current_timestamp

# When cgroup memory is not constrained, the limit value will be a high value close to 2^63 - 1,
# Ex: AWS ECS returns 2^63 - 2^10 = 9223372036854000000
UNCONSTRAINED_CGROUP_MEMORY_LIMIT = 9223372036854000000


def compatible_cgroup_version_detected() -> bool:
    return _retrieve_cgroup_version(logger=None) is not None


def cpu_usage_path_cgroup_v1() -> str:
    """Path to the cgroup file containing the CPU time in nanoseconds.

    We use cgroup files instead of the psutil library because psutil uses the host machine's CPU allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_CPU_USAGE_PATH", "/sys/fs/cgroup/cpuacct/cpuacct.usage")


def cpu_stat_path_cgroup_v2() -> str:
    """Path to the cgroup file containing current CPU stats in microseconds.

    We use cgroup files instead of the psutil library because psutil uses the host machine's CPU allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_CPU_STAT_PATH", "/sys/fs/cgroup/cpu.stat")


def cpu_max_path_cgroup_v2() -> str:
    """Path to the cgroup file containing the maximum CPU quota per period in microseconds.

    We use cgroup files instead of the psutil library because psutil uses the host machine's CPU allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_CPU_MAX_PATH", "/sys/fs/cgroup/cpu.max")


def cpu_cfs_quota_us_path() -> str:
    """Path to the cgroup file containing the CPU quota in microseconds.

    We use cgroup files instead of the psutil library because psutil uses the host machine's CPU allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_CPU_CFS_QUOTA_US_PATH", "/sys/fs/cgroup/cpu/cpu.cfs_quota_us")


def cpu_cfs_period_us_path() -> str:
    """Path to the cgroup file containing the CPU period in microseconds.

    We use cgroup files instead of the psutil library because psutil uses the host machine's CPU allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_CPU_CFS_PERIOD_US_PATH", "/sys/fs/cgroup/cpu/cpu.cfs_period_us")


def cpu_shares_path() -> str:
    """Path to the cgroup file containing the CPU shares.

    We use cgroup files instead of the psutil library because psutil uses the host machine's CPU allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_CPU_SHARES_PATH", "/sys/fs/cgroup/cpu/cpu.shares")


def cpu_info_path() -> str:
    """Path to the file containing the number of cores allocated to the container."""
    return os.getenv("DAGSTER_CPU_INFO_PATH", "/proc/cpuinfo")


def memory_usage_path_cgroup_v1() -> str:
    """Path to the cgroup file containing the memory usage in bytes.

    We use cgroup files instead of the psutil library because psutil uses the host machine's memory allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_MEMORY_USAGE_PATH_V1", "/sys/fs/cgroup/memory/memory.usage_in_bytes")


def memory_usage_path_cgroup_v2() -> str:
    """Path to the cgroup file containing the memory usage in bytes.

    We use cgroup files instead of the psutil library because psutil uses the host machine's memory allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_MEMORY_USAGE_PATH_V2", "/sys/fs/cgroup/memory.current")


def memory_limit_path_cgroup_v1() -> str:
    """Path to the cgroup file containing the memory limit in bytes.

    We use cgroup files instead of the psutil library because psutil uses the host machine's memory allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_MEMORY_LIMIT_PATH_V1", "/sys/fs/cgroup/memory/memory.limit_in_bytes")


def memory_limit_path_cgroup_v2() -> str:
    """Path to the cgroup file containing the memory limit in bytes.

    We use cgroup files instead of the psutil library because psutil uses the host machine's memory allocation in virtualized environments like Docker, K8s, ECS, etc.
    """
    return os.getenv("DAGSTER_MEMORY_LIMIT_PATH_V2", "/sys/fs/cgroup/memory.max")


class CGroupVersion(Enum):
    V1 = "V1"
    V2 = "V2"


class ContainerUtilizationMetrics(TypedDict):
    num_allocated_cores: Optional[int]
    cpu_usage: Optional[float]  # CPU usage in seconds
    cpu_cfs_quota_us: Optional[float]  # CPU quota per period in microseconds
    cpu_cfs_period_us: Optional[float]  # CPU period in microseconds
    memory_usage: Optional[float]  # Memory usage in bytes
    memory_limit: Optional[int]  # Memory limit in bytes
    measurement_timestamp: Optional[float]
    previous_cpu_usage: Optional[float]
    previous_measurement_timestamp: Optional[float]
    cgroup_version: Optional[str]


def retrieve_containerized_utilization_metrics(
    logger: Optional[logging.Logger],
    previous_measurement_timestamp: Optional[float] = None,
    previous_cpu_usage: Optional[float] = None,
) -> ContainerUtilizationMetrics:
    """Retrieve the CPU and memory utilization metrics from cgroup and proc files."""
    cgroup_version = _retrieve_cgroup_version(logger)
    return {
        "num_allocated_cores": _retrieve_containerized_num_allocated_cores(logger),
        "cpu_usage": _retrieve_containerized_cpu_usage(logger, cgroup_version),
        "previous_cpu_usage": previous_cpu_usage,
        "previous_measurement_timestamp": previous_measurement_timestamp,
        "cpu_cfs_quota_us": _retrieve_containerized_cpu_cfs_quota_us(logger, cgroup_version),
        "cpu_cfs_period_us": _retrieve_containerized_cpu_cfs_period_us(logger, cgroup_version),
        "memory_usage": _retrieve_containerized_memory_usage(logger, cgroup_version),
        "memory_limit": _retrieve_containerized_memory_limit(logger, cgroup_version),
        "measurement_timestamp": get_current_timestamp(),
        "cgroup_version": cgroup_version.value if cgroup_version else None,
    }


def _retrieve_cgroup_version(logger: Optional[logging.Logger]) -> Optional[CGroupVersion]:
    try:
        # Run the stat command in a subprocess and read the result.
        status = os.popen("stat -fc %T /sys/fs/cgroup/").read().strip()
        if status == "cgroup2fs":
            return CGroupVersion.V2
        elif status == "tmpfs":
            return CGroupVersion.V1
        else:
            return None
    except Exception as e:
        if logger:
            logger.info(f"No cgroup version found: {e}")
        return None


def _retrieve_containerized_cpu_usage(
    logger: Optional[logging.Logger], cgroup_version: Optional[CGroupVersion]
) -> Optional[float]:
    """Retrieve the CPU time in seconds from the cgroup file."""
    if cgroup_version == CGroupVersion.V1:
        return _retrieve_containerized_cpu_usage_v1(logger)
    elif cgroup_version == CGroupVersion.V2:
        return _retrieve_containerized_cpu_usage_v2(logger)
    else:
        return None


def _retrieve_containerized_cpu_usage_v1(logger: Optional[logging.Logger]) -> Optional[float]:
    try:
        with open(cpu_usage_path_cgroup_v1()) as f:
            return float(f.read()) / 1e9  # Cpuacct.usage is in nanoseconds
    except Exception as e:
        if logger:
            logger.error(f"Failed to retrieve CPU time from cgroup: {e}")
        return None


def _retrieve_containerized_cpu_usage_v2(logger: Optional[logging.Logger]) -> Optional[float]:
    try:
        with open(cpu_stat_path_cgroup_v2()) as f:
            lines = f.readlines()
            for line in lines:
                if line.startswith("usage_usec"):
                    return float(line.split()[1]) / 1e6  # Cpu.stat usage_usec is in microseconds
            return None
    except Exception as e:
        if logger:
            logger.error(f"Failed to retrieve CPU time from cgroup: {e}")
        return None


def _retrieve_containerized_num_allocated_cores(logger: Optional[logging.Logger]) -> Optional[int]:
    """Retrieve the number of cores from the /proc/cpuinfo file."""
    try:
        with open(cpu_info_path()) as f:
            return len([line for line in f if line.startswith("processor")])
    except Exception as e:
        if logger:
            logger.error(f"Failed to retrieve number of cores from /proc/cpuinfo: {e}")
        return None


def _retrieve_containerized_memory_usage(
    logger: Optional[logging.Logger], cgroup_version: Optional[CGroupVersion]
) -> Optional[int]:
    """Retrieve the memory usage in bytes from the cgroup file."""
    if cgroup_version == CGroupVersion.V1:
        return _retrieve_containerized_memory_usage_v1(logger)
    elif cgroup_version == CGroupVersion.V2:
        return _retrieve_containerized_memory_usage_v2(logger)
    else:
        return None


def _retrieve_containerized_memory_usage_v1(logger: Optional[logging.Logger]) -> Optional[int]:
    try:
        with open(memory_usage_path_cgroup_v1()) as f:
            return int(f.read())
    except Exception as e:
        if logger:
            logger.error(f"Failed to retrieve memory usage from cgroup: {e}")
        return None


def _retrieve_containerized_memory_usage_v2(logger: Optional[logging.Logger]) -> Optional[int]:
    try:
        with open(memory_usage_path_cgroup_v2()) as f:
            return int(f.read())
    except Exception as e:
        if logger:
            logger.error(f"Failed to retrieve memory usage from cgroup: {e}")
        return None


def _retrieve_containerized_memory_limit(
    logger: Optional[logging.Logger], cgroup_version: Optional[CGroupVersion]
) -> Optional[int]:
    """Retrieve the memory limit in bytes from the cgroup file."""
    if cgroup_version == CGroupVersion.V1:
        return _retrieve_containerized_memory_limit_v1(logger)
    elif cgroup_version == CGroupVersion.V2:
        return _retrieve_containerized_memory_limit_v2(logger)
    else:
        return None


def _retrieve_containerized_memory_limit_v1(logger: Optional[logging.Logger]) -> Optional[int]:
    try:
        with open(memory_limit_path_cgroup_v1()) as f:
            return int(f.read())
    except:
        if logger:
            logger.exception("Failed to retrieve memory limit from cgroup")
        return None


def _retrieve_containerized_memory_limit_v2(logger: Optional[logging.Logger]) -> Optional[int]:
    try:
        with open(memory_limit_path_cgroup_v2()) as f:
            return int(f.read())
    except:
        if logger:
            logger.exception(
                "Failed to retrieve memory limit from cgroup. There may be no limit set on the container."
            )
        return None


def _retrieve_containerized_cpu_cfs_period_us(
    logger: Optional[logging.Logger], cgroup_version: Optional[CGroupVersion]
) -> Optional[float]:
    """Retrieve the CPU period in microseconds from the cgroup file."""
    if cgroup_version == CGroupVersion.V1:
        return _retrieve_containerized_cpu_cfs_period_us_v1(logger)
    elif cgroup_version == CGroupVersion.V2:
        return _retrieve_containerized_cpu_cfs_period_us_v2(logger)
    else:
        return None


def _retrieve_containerized_cpu_cfs_period_us_v1(
    logger: Optional[logging.Logger],
) -> Optional[float]:
    try:
        with open(cpu_cfs_period_us_path()) as f:
            return float(f.read())
    except:
        if logger:
            logger.exception("Failed to retrieve CPU period from cgroup")
        return None


def _retrieve_containerized_cpu_cfs_period_us_v2(
    logger: Optional[logging.Logger],
) -> Optional[float]:
    # We can retrieve period information from the cpu.max file. The file is in the format $MAX $PERIOD and is only one line.
    try:
        with open(cpu_max_path_cgroup_v2()) as f:
            line = f.readline()
            return float(line.split()[1])
    except:
        if logger:
            logger.exception("Failed to retrieve CPU period from cgroup")
        return None


def _retrieve_containerized_cpu_cfs_quota_us(
    logger: Optional[logging.Logger], cgroup_version: Optional[CGroupVersion]
) -> Optional[float]:
    """Retrieve the CPU quota in microseconds from the cgroup file."""
    if cgroup_version == CGroupVersion.V1:
        return _retrieve_containerized_cpu_cfs_quota_us_v1(logger)
    elif cgroup_version == CGroupVersion.V2:
        return _retrieve_containerized_cpu_cfs_quota_us_v2(logger)
    else:
        return None


def _retrieve_containerized_cpu_cfs_quota_us_v1(
    logger: Optional[logging.Logger],
) -> Optional[float]:
    try:
        with open(cpu_cfs_quota_us_path()) as f:
            return float(f.read())
    except:
        if logger:
            logger.exception("Failed to retrieve CPU quota from cgroup")
        return None


def _retrieve_containerized_cpu_cfs_quota_us_v2(
    logger: Optional[logging.Logger],
) -> Optional[float]:
    # We can retrieve quota information from the cpu.max file. The file is in the format $MAX $PERIOD .
    try:
        with open(cpu_max_path_cgroup_v2()) as f:
            line = f.readline()
            return float(line.split()[0])
    except:
        if logger:
            logger.exception(
                "Failed to retrieve CPU quota from cgroup. There might be not limit set on the container."
            )
        return None



import inspect
from collections.abc import AsyncIterator, Awaitable, Iterator, Mapping, Sequence
from functools import wraps
from typing import Any, Callable, Optional, Union, cast

from typing_extensions import get_args

from dagster._config.pythonic_config import Config
from dagster._core.definitions import (
    AssetCheckResult,
    AssetMaterialization,
    DynamicOutput,
    ExpectationResult,
    Output,
    OutputDefinition,
)
from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction
from dagster._core.definitions.input import InputDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.result import AssetResult, ObserveResult
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.execution.context.compute import ExecutionContextTypes
from dagster._core.types.dagster_type import DagsterTypeKind, is_generic_output_annotation
from dagster._utils import is_named_tuple_instance
from dagster._utils.warnings import disable_dagster_warnings


def create_op_compute_wrapper(
    op_def: OpDefinition,
) -> Callable[[ExecutionContextTypes, Mapping[str, InputDefinition]], Any]:
    compute_fn = cast(DecoratedOpFunction, op_def.compute_fn)
    fn = compute_fn.decorated_fn
    input_defs = op_def.input_defs
    output_defs = op_def.output_defs
    context_arg_provided = compute_fn.has_context_arg()
    config_arg_cls = compute_fn.get_config_arg().annotation if compute_fn.has_config_arg() else None
    resource_arg_mapping = {arg.name: arg.name for arg in compute_fn.get_resource_args()}

    input_names = [
        input_def.name
        for input_def in input_defs
        if not input_def.dagster_type.kind == DagsterTypeKind.NOTHING
    ]

    @wraps(fn)
    def compute(
        context: ExecutionContextTypes,
        input_defs: Mapping[str, InputDefinition],
    ) -> Union[Iterator[Output], AsyncIterator[Output]]:
        kwargs = {}
        for input_name in input_names:
            kwargs[input_name] = input_defs[input_name]

        if (
            inspect.isgeneratorfunction(fn)
            or inspect.isasyncgenfunction(fn)
            or inspect.iscoroutinefunction(fn)
        ):
            # safe to execute the function, as doing so will not immediately execute user code
            result = invoke_compute_fn(
                fn, context, kwargs, context_arg_provided, config_arg_cls, resource_arg_mapping
            )
            if inspect.iscoroutine(result):
                return _coerce_async_op_to_async_gen(result, context, output_defs)
            # already a generator
            return result
        else:
            # we have a regular function, do not execute it before we are in an iterator
            # (as we want all potential failures to happen inside iterators)
            return _coerce_op_compute_fn_to_iterator(
                fn,
                output_defs,
                context,
                context_arg_provided,
                kwargs,
                config_arg_cls,
                resource_arg_mapping,
            )

    return compute


async def _coerce_async_op_to_async_gen(
    awaitable: Awaitable[Any],
    context: ExecutionContextTypes,
    output_defs: Sequence[OutputDefinition],
) -> AsyncIterator[Any]:
    result = await awaitable
    for event in validate_and_coerce_op_result_to_iterator(result, context, output_defs):
        yield event


def invoke_compute_fn(
    fn: Callable[..., Any],
    context: ExecutionContextTypes,
    kwargs: Mapping[str, Any],
    context_arg_provided: bool,
    config_arg_cls: Optional[type[Config]],
    resource_args: Optional[dict[str, str]] = None,
) -> Any:
    args_to_pass = {**kwargs}
    if config_arg_cls:
        # config_arg_cls is either a Config class or a primitive type
        if issubclass(config_arg_cls, Config):
            to_pass = config_arg_cls._get_non_default_public_field_values_cls(  # noqa: SLF001
                context.op_execution_context.op_config
            )
            args_to_pass["config"] = config_arg_cls(**to_pass)
        else:
            args_to_pass["config"] = context.op_execution_context.op_config
    if resource_args:
        for resource_name, arg_name in resource_args.items():
            args_to_pass[arg_name] = context.resources.original_resource_dict[resource_name]

    return fn(context, **args_to_pass) if context_arg_provided else fn(**args_to_pass)


def _coerce_op_compute_fn_to_iterator(
    fn,
    output_defs,
    context: ExecutionContextTypes,
    context_arg_provided,
    kwargs,
    config_arg_class,
    resource_arg_mapping,
):
    result = invoke_compute_fn(
        fn, context, kwargs, context_arg_provided, config_arg_class, resource_arg_mapping
    )
    yield from validate_and_coerce_op_result_to_iterator(result, context, output_defs)


def _zip_and_iterate_op_result(
    result: Any, context: ExecutionContextTypes, output_defs: Sequence[OutputDefinition]
) -> Iterator[tuple[int, Any, OutputDefinition]]:
    # Filtering the expected output defs here is an unfortunate temporary solution to deal with the
    # change in expected outputs that occurs as a result of putting `AssetCheckResults` onto
    # `MaterializeResults`. Prior to this, `AssetCheckResults` were yielded/returned directly, and
    # thus were expected to always be included in the result tuple. Thus we need to remove them from
    # the expected output defs if they have been included indirectly via embedding in a
    # `MaterializeResult`.
    #
    # A better solution is surely possible here in a future refactor. The major complicating element
    # is the conversion of MaterializeResult into Output, which currently happens in
    # execute_step.py and leverages job_def.asset_layer. There is difficulty in moving that logic up
    # to here because we can't rely on the presence of asset layer here, since the present code path
    # is used by direct invocation. Probably the solution is to expose an asset layer on the
    # invocation context.
    expected_return_outputs = _filter_expected_output_defs(result, context, output_defs)
    if len(expected_return_outputs) > 1:
        result = _validate_multi_return(context, result, expected_return_outputs)
        for position, (output_def, element) in enumerate(zip(expected_return_outputs, result)):
            yield position, output_def, element
    else:
        yield 0, output_defs[0], result


# Filter out output_defs corresponding to asset check results that already exist on a
# MaterializeResult.
def _filter_expected_output_defs(
    result: Any, context: ExecutionContextTypes, output_defs: Sequence[OutputDefinition]
) -> Sequence[OutputDefinition]:
    result_tuple = (
        (result,) if not isinstance(result, tuple) or is_named_tuple_instance(result) else result
    )
    asset_results = [x for x in result_tuple if isinstance(x, AssetResult)]

    if not asset_results:
        return output_defs

    check_names_by_asset_key = {}
    for check_key in context.op_execution_context.selected_asset_check_keys:
        if check_key.asset_key not in check_names_by_asset_key:
            check_names_by_asset_key[check_key.asset_key] = []
        check_names_by_asset_key[check_key.asset_key].append(check_key.name)

    remove_outputs = []
    for asset_result in asset_results:
        for check_result in asset_result.check_results:
            remove_outputs.append(
                context.op_execution_context.assets_def.get_output_name_for_asset_check_key(
                    check_result.resolve_target_check_key(check_names_by_asset_key)
                )
            )

    return [out for out in output_defs if out.name not in remove_outputs]


def _validate_multi_return(
    context: ExecutionContextTypes,
    result: Any,
    output_defs: Sequence[OutputDefinition],
) -> Any:
    # special cases for implicit/explicit returned None
    if result is None:
        # extrapolate None -> (None, None, ...) when appropriate
        if all(
            output_def.dagster_type.is_nothing and output_def.is_required
            for output_def in output_defs
        ):
            return [None for _ in output_defs]

    # When returning from an op with multiple outputs, the returned object must be a tuple of the same length as the number of outputs. At the time of the op's construction, we verify that a provided annotation is a tuple with the same length as the number of outputs, so if the result matches the number of output defs on the op, it will transitively also match the annotation.
    if not isinstance(result, tuple):
        raise DagsterInvariantViolationError(
            f"{context.describe_op()} has multiple outputs, but only one "
            f"output was returned of type {type(result)}. When using "
            "multiple outputs, either yield each output, or return a tuple "
            "containing a value for each output. Check out the "
            "documentation on outputs for more: "
            "https://legacy-docs.dagster.io/concepts/ops-jobs-graphs/ops#outputs."
        )
    output_tuple = cast(tuple, result)
    if not len(output_tuple) == len(output_defs):
        raise DagsterInvariantViolationError(
            "Length mismatch between returned tuple of outputs and number of "
            f"output defs on {context.describe_op()}. Output tuple has "
            f"{len(output_tuple)} outputs, while "
            f"{context.op_def.node_type_str} has {len(output_defs)} outputs."
        )
    return result


def _get_annotation_for_output_position(
    position: int, op_def: OpDefinition, output_defs: Sequence[OutputDefinition]
) -> Any:
    if op_def.is_from_decorator():
        if len(output_defs) > 1:
            annotation_subitems = get_args(op_def.get_output_annotation())
            if len(annotation_subitems) == len(output_defs):
                return annotation_subitems[position]
        else:
            return op_def.get_output_annotation()

    return inspect.Parameter.empty


def _check_output_object_name(
    output: Union[DynamicOutput, Output], output_def: OutputDefinition, position: int
) -> None:
    from dagster._core.definitions.events import DEFAULT_OUTPUT

    if not output.output_name == DEFAULT_OUTPUT and not output.output_name == output_def.name:
        raise DagsterInvariantViolationError(
            f"Bad state: Output was explicitly named '{output.output_name}', "
            "which does not match the output definition specified for position "
            f"{position}: '{output_def.name}'."
        )


def validate_and_coerce_op_result_to_iterator(
    result: Any,
    context: ExecutionContextTypes,
    output_defs: Sequence[OutputDefinition],
) -> Iterator[Any]:
    if inspect.isgenerator(result):
        # this happens when a user explicitly returns a generator in the op
        yield from result
    elif isinstance(result, (AssetMaterialization, ExpectationResult)):
        raise DagsterInvariantViolationError(
            f"Error in {context.describe_op()}: If you are "
            "returning an AssetMaterialization "
            "or an ExpectationResult from "
            f"{context.op_def.node_type_str} you must yield them "
            "directly, or log them using the context.log_event method to avoid "
            "ambiguity with an implied result from returning a "
            "value. Check out the docs on logging events here: "
            "https://legacy-docs.dagster.io/concepts/ops-jobs-graphs/op-events#op-events-and-exceptions"
        )
    # These don't correspond to output defs so pass them through
    elif isinstance(result, (AssetCheckResult, ObserveResult)):
        yield result
    elif result is not None and not output_defs:
        raise DagsterInvariantViolationError(
            f"Error in {context.describe_op()}: Unexpectedly returned output of type"
            f" {type(result)}. {context.op_def.node_type_str.capitalize()} is explicitly defined to"
            " return no results."
        )
    # `requires_typed_event_stream` is a mode where we require users to return/yield exactly the
    # results that will be registered in the instance, without additional fancy inference (like
    # wrapping `None` in an `Output`). We therefore skip any return-specific validation for this
    # mode and treat returned values as if they were yielded.
    elif output_defs and context.op_execution_context.requires_typed_event_stream:
        # If nothing was returned, treat it as an empty tuple instead of a `(None,)`.
        # This is important for delivering the correct error message when an output is missing.
        if result is None:
            result_tuple = tuple()
        elif not isinstance(result, tuple) or is_named_tuple_instance(result):
            result_tuple = (result,)
        else:
            result_tuple = result
        yield from result_tuple
    elif output_defs:
        for position, output_def, element in _zip_and_iterate_op_result(
            result, context, output_defs
        ):
            annotation = _get_annotation_for_output_position(position, context.op_def, output_defs)
            if output_def.is_dynamic:
                if not isinstance(element, list):
                    raise DagsterInvariantViolationError(
                        f"Error with output for {context.describe_op()}: "
                        f"dynamic output '{output_def.name}' expected a list of "
                        "DynamicOutput objects, but instead received instead an "
                        f"object of type {type(element)}."
                    )
                for item in element:
                    if not isinstance(item, DynamicOutput):
                        raise DagsterInvariantViolationError(
                            f"Error with output for {context.describe_op()}: "
                            f"dynamic output '{output_def.name}' at position {position} expected a "
                            "list of DynamicOutput objects, but received an "
                            f"item with type {type(item)}."
                        )
                    dynamic_output = cast(DynamicOutput, item)
                    _check_output_object_name(dynamic_output, output_def, position)

                    with disable_dagster_warnings():
                        yield DynamicOutput(
                            output_name=output_def.name,
                            value=dynamic_output.value,
                            mapping_key=dynamic_output.mapping_key,
                            metadata=dynamic_output.metadata,
                        )
            elif isinstance(element, AssetResult):
                yield element  # coerced in to Output in outer iterator
            elif isinstance(element, Output):
                if annotation != inspect.Parameter.empty and not is_generic_output_annotation(
                    annotation
                ):
                    raise DagsterInvariantViolationError(
                        f"Error with output for {context.describe_op()}: received Output object for"
                        f" output '{output_def.name}' which does not have an Output annotation."
                        f" Annotation has type {annotation}."
                    )
                _check_output_object_name(element, output_def, position)

                with disable_dagster_warnings():
                    output = Output(
                        output_name=output_def.name,
                        value=element.value,
                        metadata=element.metadata,
                        data_version=element.data_version,
                        tags=element.tags,
                    )

                yield output

            else:
                # If annotation indicates a generic output annotation, and an
                # output object was not received, throw an error.
                if is_generic_output_annotation(annotation):
                    raise DagsterInvariantViolationError(
                        f"Error with output for {context.describe_op()}: output "
                        f"'{output_def.name}' has generic output annotation, "
                        "but did not receive an Output object for this output. "
                        f"Received instead an object of type {type(element)}."
                    )
                if result is None and output_def.is_required is False:
                    context.log.warning(
                        'Value "None" returned for non-required output '
                        f'"{output_def.name}" of {context.describe_op()}. '
                        "This value will be passed to downstream "
                        f"{context.op_def.node_type_str}s. For conditional "
                        "execution, results must be yielded: "
                        "https://legacy-docs.dagster.io/concepts/ops-jobs-graphs/graphs#with-conditional-branching"
                    )
                yield Output(output_name=output_def.name, value=element)



from abc import ABC, abstractproperty
from enum import Enum
from typing import NamedTuple, Optional

from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionEvaluation,
    AutomationConditionEvaluationWithRunIds,
    AutomationConditionNodeSnapshot,
    HistoricalAllPartitionsSubsetSentinel,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MetadataMapping, MetadataValue
from dagster._serdes.serdes import (
    NamedTupleSerializer,
    UnpackContext,
    UnpackedValue,
    WhitelistMap,
    whitelist_for_serdes,
)


@whitelist_for_serdes
class AutoMaterializeDecisionType(Enum):
    """Represents the set of results of the auto-materialize logic.

    MATERIALIZE: The asset should be materialized by a run kicked off on this tick
    SKIP: The asset should not be materialized by a run kicked off on this tick, because future
        ticks are expected to materialize it.
    DISCARD: The asset should not be materialized by a run kicked off on this tick, but future
        ticks are not expected to materialize it.
    """

    MATERIALIZE = "MATERIALIZE"
    SKIP = "SKIP"
    DISCARD = "DISCARD"


class AutoMaterializeRuleEvaluationData(ABC):
    @abstractproperty
    def metadata(self) -> MetadataMapping:
        raise NotImplementedError()

    @property
    def frozen_metadata(self) -> frozenset[tuple[str, MetadataValue]]:
        return frozenset(self.metadata.items())


@whitelist_for_serdes
class TextRuleEvaluationData(
    AutoMaterializeRuleEvaluationData,
    NamedTuple("_TextRuleEvaluationData", [("text", str)]),
):
    @property
    def metadata(self) -> MetadataMapping:
        return {"text": MetadataValue.text(self.text)}


@whitelist_for_serdes
class ParentUpdatedRuleEvaluationData(
    AutoMaterializeRuleEvaluationData,
    NamedTuple(
        "_ParentUpdatedRuleEvaluationData",
        [
            ("updated_asset_keys", frozenset[AssetKey]),
            ("will_update_asset_keys", frozenset[AssetKey]),
        ],
    ),
):
    @property
    def metadata(self) -> MetadataMapping:
        return {
            **{
                f"updated_parent_{i+1}": MetadataValue.asset(k)
                for i, k in enumerate(sorted(self.updated_asset_keys))
            },
            **{
                f"will_update_parent_{i+1}": MetadataValue.asset(k)
                for i, k in enumerate(sorted(self.will_update_asset_keys))
            },
        }


@whitelist_for_serdes
class WaitingOnAssetsRuleEvaluationData(
    AutoMaterializeRuleEvaluationData,
    NamedTuple(
        "_WaitingOnParentRuleEvaluationData",
        [("waiting_on_asset_keys", frozenset[AssetKey])],
    ),
):
    @property
    def metadata(self) -> MetadataMapping:
        return {
            **{
                f"waiting_on_ancestor_{i+1}": MetadataValue.asset(k)
                for i, k in enumerate(sorted(self.waiting_on_asset_keys))
            },
        }


# BACKCOMPAT GRAVEYARD


class BackcompatNullSerializer(NamedTupleSerializer):
    """Unpacks an arbitrary object into None."""

    def unpack(
        self,
        unpacked_dict: dict[str, UnpackedValue],
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> None:
        return None


class BackcompatAutoMaterializeAssetEvaluationSerializer(NamedTupleSerializer):
    """Unpacks the legacy AutoMaterializeAssetEvaluation class into a completely empty
    AutomationConditionEvaluationWithRunIds.
    """

    def unpack(
        self,
        unpacked_dict: dict[str, UnpackedValue],
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> "AutomationConditionEvaluationWithRunIds":
        return AutomationConditionEvaluationWithRunIds(
            evaluation=AutomationConditionEvaluation(
                condition_snapshot=AutomationConditionNodeSnapshot("", "", "", None, None),
                start_timestamp=None,
                end_timestamp=None,
                true_subset=SerializableEntitySubset(key=AssetKey("unknown"), value=False),
                candidate_subset=HistoricalAllPartitionsSubsetSentinel(),
                subsets_with_metadata=[],
                child_evaluations=[],
            ),
            run_ids=frozenset(),
        )


@whitelist_for_serdes(serializer=BackcompatAutoMaterializeAssetEvaluationSerializer)
class AutoMaterializeAssetEvaluation(NamedTuple): ...


@whitelist_for_serdes(serializer=BackcompatNullSerializer)
class AutoMaterializeRuleSnapshot(NamedTuple):
    class_name: str
    description: str
    decision_type: AutoMaterializeDecisionType


@whitelist_for_serdes(serializer=BackcompatNullSerializer)
class AutoMaterializeRuleEvaluation(NamedTuple):
    rule_snapshot: AutoMaterializeRuleSnapshot
    evaluation_data: Optional[AutoMaterializeRuleEvaluationData]
